import pytest
from graphene import Context, Enum as GrapheneEnum, ObjectType, Schema
from sqlalchemy.types import Enum as SQLAlchemyEnum

from alchql import gql_types
from alchql.fields import BatchSQLAlchemyConnectionField, SQLAlchemyConnectionField
from alchql.middlewares import LoaderMiddleware
from alchql.node import AsyncNode
from alchql.registry import Registry
from alchql.types import SQLAlchemyObjectType
from alchql.utils import EnumValue
from .models import Editor, Pet


def test_register_object_type():
    reg = Registry()

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            registry = reg

    reg.register(PetType)
    assert reg.get_type_for_model(Pet) is PetType


def test_register_incorrect_object_type():
    reg = Registry()

    class Spam:
        pass

    re_err = "Expected SQLAlchemyObjectType, but got: .*Spam"
    with pytest.raises(TypeError, match=re_err):
        reg.register(Spam)


def test_register_orm_field():
    reg = Registry()

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            registry = reg

    reg.register_orm_field(PetType, "name", Pet.name)
    assert reg.get_orm_field_for_graphene_field(PetType, "name") is Pet.name


def test_register_orm_field_incorrect_types():
    reg = Registry()

    class Spam:
        pass

    re_err = "Expected SQLAlchemyObjectType, but got: .*Spam"
    with pytest.raises(TypeError, match=re_err):
        reg.register_orm_field(Spam, "name", Pet.name)

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            registry = reg

    re_err = "Expected a field name, but got: .*Spam"
    with pytest.raises(TypeError, match=re_err):
        reg.register_orm_field(PetType, Spam, Pet.name)


def test_register_enum():
    reg = Registry()

    sa_enum = SQLAlchemyEnum("cat", "dog")
    graphene_enum = GrapheneEnum("PetKind", [("CAT", 1), ("DOG", 2)])

    reg.register_enum(sa_enum, graphene_enum)
    assert reg.get_graphene_enum_for_sa_enum(sa_enum) is graphene_enum


def test_register_enum_incorrect_types():
    reg = Registry()

    sa_enum = SQLAlchemyEnum("cat", "dog")
    graphene_enum = GrapheneEnum("PetKind", [("CAT", 1), ("DOG", 2)])

    re_err = r"Expected Graphene Enum, but got: Enum\('cat', 'dog'\)"
    with pytest.raises(TypeError, match=re_err):
        reg.register_enum(sa_enum, sa_enum)

    re_err = r"Expected SQLAlchemyEnumType, but got: .*PetKind.*"
    with pytest.raises(TypeError, match=re_err):
        reg.register_enum(graphene_enum, graphene_enum)


def test_register_sort_enum():
    reg = Registry()

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            registry = reg

    sort_enum = GrapheneEnum(
        "PetSort",
        [("ID", EnumValue("id", Pet.id)), ("NAME", EnumValue("name", Pet.name))],
    )

    reg.register_sort_enum(PetType, sort_enum)
    assert reg.get_sort_enum_for_object_type(PetType) is sort_enum


def test_register_sort_enum_incorrect_types():
    reg = Registry()

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            registry = reg

    sort_enum = GrapheneEnum(
        "PetSort",
        [("ID", EnumValue("id", Pet.id)), ("NAME", EnumValue("name", Pet.name))],
    )

    re_err = r"Expected SQLAlchemyObjectType, but got: .*PetSort.*"
    with pytest.raises(TypeError, match=re_err):
        reg.register_sort_enum(sort_enum, sort_enum)

    re_err = r"Expected Graphene Enum, but got: .*PetType.*"
    with pytest.raises(TypeError, match=re_err):
        reg.register_sort_enum(PetType, PetType)


@pytest.mark.asyncio
async def test_sqlalchemy_redefine_field4(session):
    class EditorType(SQLAlchemyObjectType):
        class Meta:
            model = Editor
            connection_field_factory = BatchSQLAlchemyConnectionField.from_relationship
            interfaces = (AsyncNode,)

        pname = gql_types.String(model_field=Editor.name)

    class AnotherEditorType(SQLAlchemyObjectType):
        class Meta:
            model = Editor
            connection_field_factory = BatchSQLAlchemyConnectionField.from_relationship
            interfaces = (AsyncNode,)

        dname = gql_types.String(model_field=Editor.name)

    class Query(ObjectType):
        editor = SQLAlchemyConnectionField(EditorType.connection, sort=None)
        another_editor = SQLAlchemyConnectionField(
            AnotherEditorType.connection, sort=None
        )

    editor = Editor(name="first_name")
    session.add(editor)
    await session.commit()

    schema = Schema(query=Query, types=[EditorType, AnotherEditorType])
    result = await schema.execute_async(
        """
        query {
            editor {
                edges {
                    node {
                        pname
                    }
                }
            }
            anotherEditor {
                edges {
                    node {
                        dname
                    }
                }
            }
        }
        """,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Editor]),
        ],
    )

    assert not result.errors
    assert result.data["editor"]["edges"][0]["node"]["pname"] == "first_name"
    assert result.data["anotherEditor"]["edges"][0]["node"]["dname"] == "first_name"



import enum
from functools import partial
from inspect import isawaitable
from typing import Type

import graphene
import sqlalchemy as sa
from graphene.relay import Connection, ConnectionField
from graphene.types import ResolveInfo
from graphene.types.utils import get_type
from sqlalchemy import ForeignKey
from sqlalchemy.orm import DeclarativeMeta, InstrumentedAttribute, RelationshipProperty
from sqlalchemy.sql.elements import Label

from .batching import get_batch_resolver, get_fk_resolver_reverse
from .connection.from_array_slice import connection_from_array_slice
from .connection.from_query import connection_from_query
from .consts import OPERATORS_MAPPING, OP_EQ, OP_IN
from .query_helper import QueryHelper
from .registry import Registry, get_global_registry
from .sqlalchemy_converter import convert_sqlalchemy_type
from .utils import EnumValue, FilterItem, GlobalFilters, get_query


class UnsortedSQLAlchemyConnectionField(ConnectionField):
    @property
    def type(self):
        from .types import SQLAlchemyObjectType

        type_ = super(ConnectionField, self).type
        nullable_type = get_nullable_type(type_)
        if issubclass(nullable_type, Connection):
            return type_

        if not issubclass(nullable_type, SQLAlchemyObjectType):
            raise AssertionError(
                f"SQLALchemyConnectionField only accepts SQLAlchemyObjectType types, "
                f"not {nullable_type.__name__}"
            )
        if not nullable_type.connection:
            raise AssertionError(
                f"The type {nullable_type.__name__} doesn't have a connection"
            )
        if type_ != nullable_type:
            raise AssertionError(
                "Passing a SQLAlchemyObjectType instance is deprecated. "
                "Pass the connection type instead accessible via SQLAlchemyObjectType.connection"
            )
        return nullable_type.connection

    @property
    def model(self):
        return get_nullable_type(self.type)._meta.node._meta.model

    @classmethod
    async def get_query(
        cls, model: Type[DeclarativeMeta], info: ResolveInfo, cls_name=None, **args
    ):
        return get_query(model, info, cls_name)

    @classmethod
    async def resolve_connection(
        cls,
        connection_type,
        model: Type[DeclarativeMeta],
        info: ResolveInfo,
        args,
        resolved,
    ):
        if resolved is None:
            edge_type = connection_type.Edge
            node_type = edge_type.node.type

            query = await cls.get_query(
                model=model,
                info=info,
                cls_name=node_type.__name__,
                **args,
            )

            connection = await connection_from_query(
                query,
                info=info,
                model=model,
                args=args,
                connection_type=connection_type,
            )
        else:
            if isawaitable(resolved):
                resolved = await resolved

            connection = connection_from_array_slice(
                array_slice=resolved,
                args=args,
                connection_type=connection_type,
            )

            if hasattr(connection, "total_count"):
                connection.total_count = len(resolved)
        return connection

    @classmethod
    async def connection_resolver(
        cls,
        resolver,
        connection_type,
        model: Type[DeclarativeMeta],
        root,
        info: ResolveInfo,
        **args,
    ):
        types = getattr(info.context, "object_types", {})
        types[info.field_name] = connection_type.Edge.node.type
        setattr(info.context, "object_types", types)
        resolved = resolver(root, info, **args)

        on_resolve = partial(cls.resolve_connection, connection_type, model, info, args)
        result = on_resolve(resolved)

        if isawaitable(result):
            return await result
        return result

    def wrap_resolve(self, parent_resolver):
        return partial(
            self.connection_resolver,
            parent_resolver,
            get_nullable_type(self.type),
            self.model,
        )


class SQLAlchemyConnectionField(UnsortedSQLAlchemyConnectionField):
    def __init__(self, type_, *args, **kwargs):
        nullable_type = get_nullable_type(type_)
        if "sort" not in kwargs and issubclass(nullable_type, Connection):
            # Let super class raise if type is not a Connection
            try:
                kwargs.setdefault("sort", nullable_type.Edge.node._type.sort_argument())
            except (AttributeError, TypeError):
                raise TypeError(
                    f"Cannot create sort argument for {nullable_type.__name__}. "
                    'A model is required. Set the "sort" argument to None '
                    "to disabling the creation of the sort query argument"
                )
        elif "sort" in kwargs and kwargs["sort"] is None:
            del kwargs["sort"]

        super().__init__(type_, *args, **kwargs)

    @classmethod
    async def get_query(
        cls,
        model: Type[DeclarativeMeta],
        info: ResolveInfo,
        sort=None,
        cls_name=None,
        **args,
    ):
        query = get_query(model, info, cls_name)
        if sort is not None:
            if not isinstance(sort, list):
                sort = [sort]
            sort_args = []
            # ensure consistent handling of graphene Enums, enum values and
            # plain strings
            for item in sort:
                if isinstance(item, enum.Enum):
                    sort_args.append(item.value)
                elif isinstance(item, EnumValue):
                    sort_args.append(item.value)
                else:
                    sort_args.append(item)
            query = query.order_by(*sort_args)
        return query


class FilterConnectionField(SQLAlchemyConnectionField):
    def __init__(self, type_, *args, **kwargs):
        type_ = get_type(type_)
        FilterConnectionField.set_filter_fields(type_, kwargs)

        if hasattr(type, "sort_argument"):
            kwargs["sort"] = type_.sort_argument()

        super().__init__(type_.connection, *args, **kwargs)

    @classmethod
    def set_filter_fields(cls, type_, kwargs):
        filters = {}
        kwargs[GlobalFilters.ID__EQ] = graphene.Argument(type_=graphene.ID)
        filters[GlobalFilters.ID__EQ] = FilterItem(
            filter_func=getattr(
                sa.inspect(type_._meta.model).primary_key[0],
                OPERATORS_MAPPING[OP_EQ][0],
            ),
            field_type=graphene.ID,
        )
        kwargs[GlobalFilters.ID__IN] = graphene.Argument(
            type_=graphene.List(of_type=graphene.ID)
        )
        filters[GlobalFilters.ID__IN] = FilterItem(
            filter_func=getattr(
                sa.inspect(type_._meta.model).primary_key[0],
                OPERATORS_MAPPING[OP_IN][0],
            ),
            field_type=graphene.List(of_type=graphene.ID),
        )

        for field, operators in getattr(type_._meta, "filter_fields", {}).items():
            if isinstance(field, str) and isinstance(operators, FilterItem):
                kwargs[field] = graphene.Argument(
                    type_=operators.field_type,
                    default_value=operators.default_value,
                    description=operators.description,
                    name=operators.name,
                    required=operators.required,
                )
                filters[field] = operators
                continue

            if isinstance(field, InstrumentedAttribute):
                field_key, field_type = cls.process_instrumented_field(field, type_)
            elif isinstance(field, Label):
                field_key, field_type = cls.process_label_field(field)
            elif isinstance(field, sa.Column):
                field_key, field_type = cls.process_column_field(field, type_)
            else:
                continue

            for operator in operators:
                if operator == OP_IN:
                    operator_field_type = graphene.List(of_type=field_type)
                else:
                    operator_field_type = field_type

                filter_name = f"{field_key}__{operator}"
                kwargs[filter_name] = graphene.Argument(type_=operator_field_type)
                filters[filter_name] = FilterItem(
                    field_type=operator_field_type,
                    filter_func=getattr(field, OPERATORS_MAPPING[operator][0]),
                    value_func=OPERATORS_MAPPING[operator][1],
                )

        setattr(type_, "parsed_filters", filters)

    @staticmethod
    def process_instrumented_field(field, type_):
        registry = get_global_registry()
        tablename = type_._meta.model.__tablename__

        if field.parent.tables[0].name == tablename:
            gql_field = type_._meta.fields.get(field.key)
            if gql_field and gql_field.name is not None:
                field_key = gql_field.name
            else:
                field_key = field.key
        else:
            field_key = f"{field.parent.tables[0].name}_{field.key}"

        field_type = convert_sqlalchemy_type(
            getattr(field, "type", None), field, registry
        )

        if field.prop.columns[0].foreign_keys:
            field_type = graphene.ID

        return field_key, field_type

    @staticmethod
    def process_label_field(field):
        registry = get_global_registry()
        field_key = field.key

        field_type = convert_sqlalchemy_type(
            getattr(field, "type", None), field, registry
        )

        if field.foreign_keys:
            field_type = graphene.ID

        return field_key, field_type

    @staticmethod
    def process_column_field(field, type_):
        registry = get_global_registry()
        tablename = type_._meta.model.__tablename__

        if field.table.name == tablename:
            gql_field = type_._meta.fields.get(field.key)
            if gql_field and gql_field.name is not None:
                field_key = gql_field.name
            else:
                field_key = field.key
        else:
            field_key = f"{field.table.name}_{field.key}"

        field_type = convert_sqlalchemy_type(
            getattr(field, "type", None), field, registry
        )

        if field.foreign_keys:
            field_type = graphene.ID

        return field_key, field_type

    @classmethod
    async def get_query(
        cls, model: Type[DeclarativeMeta], info: ResolveInfo, sort=None, **args
    ):
        object_types = getattr(info.context, "object_types", {})
        object_type = object_types[info.field_name]

        filters = QueryHelper.get_filters(info)

        select_fields = QueryHelper.get_selected_fields(info, model, object_type, sort)
        query = sa.select(*select_fields).select_from(model)

        if object_type and hasattr(object_type, "set_select_from"):
            gql_field = QueryHelper.get_current_field(info)
            query = await object_type.set_select_from(info, query, gql_field.values)

        if filters:
            query = query.where(sa.and_(*filters))

        if sort is not None:
            if not isinstance(sort, list):
                sort = [sort]
            sort_args = []
            # ensure consistent handling of graphene Enums, enum values and
            # plain strings
            for item in sort:
                if isinstance(item, (EnumValue, enum.Enum)):
                    sort_args.append(item.value)
                else:
                    sort_args.append(item)

            sort_args.extend(sa.inspect(model).primary_key)

            query = query.order_by(*sort_args)

        return query


class ModelField(graphene.Field):
    def __init__(self, *args, model_field=None, use_label=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_field = model_field
        self.use_label = use_label


class RelationModelField(ModelField):
    @staticmethod
    def get_default_resolver(gql_type, model_field, key_field=None):
        def resolver(self, info, *args, **kwargs):
            target_value = getattr(self, model_field.key)
            if target_value is None:
                return
            if key_field is not None:
                assert key_field in gql_type._keys[0]
                target_key_field = key_field
            else:
                target_key_field = gql_type._keys[0]
            return gql_type(**{target_key_field: target_value})

        return resolver

    def __init__(self, type_, *args, model_field, key_field=None, **kwargs):
        if "resolver" not in kwargs:
            kwargs["resolver"] = self.get_default_resolver(
                type_, model_field, key_field
            )
        kwargs["use_label"] = False
        super().__init__(type_, *args, model_field=model_field, **kwargs)


class BatchSQLAlchemyConnectionField(FilterConnectionField, ModelField):
    """
    This is currently experimental.
    The API and behavior may change in future versions.
    Use at your own risk.
    """

    def wrap_resolve(self, parent_resolver):
        return partial(
            self.connection_resolver,
            self.resolver,
            get_nullable_type(self.type),
            self.model,
        )

    @classmethod
    def from_relationship(cls, relationship, registry, **field_kwargs):
        model = relationship.mapper.entity
        model_type = registry.get_type_for_model(model)
        resolver = get_batch_resolver(relationship)

        if hasattr(model_type._meta, "filter_fields"):
            BatchSQLAlchemyConnectionField.set_filter_fields(model_type, field_kwargs)

        if hasattr(model_type, "sort_argument"):
            field_kwargs["sort"] = model_type.sort_argument()

        return cls(model_type, resolver=resolver, **field_kwargs)

    @classmethod
    def from_fk(cls, fk: ForeignKey, registry, **field_kwargs):
        model_type = registry.get_type_for_model(fk.constraint.table)

        if not model_type:
            return

        resolver = get_fk_resolver_reverse(fk, single=False)

        if hasattr(model_type._meta, "filter_fields"):
            BatchSQLAlchemyConnectionField.set_filter_fields(model_type, field_kwargs)

        if hasattr(model_type, "sort_argument"):
            field_kwargs["sort"] = model_type.sort_argument()

        return cls(model_type, resolver=resolver, **field_kwargs)


def default_connection_field_factory(
    relationship: RelationshipProperty,
    registry: Registry,
    **field_kwargs,
):
    model = relationship.mapper.entity
    model_type = registry.get_type_for_model(model)
    return UnsortedSQLAlchemyConnectionField(model_type, **field_kwargs)


def get_nullable_type(_type):
    if isinstance(_type, graphene.NonNull):
        return _type.of_type
    return _type



import graphene
import pytest
import sqlalchemy as sa
from graphene import Context

from .models import HairKind, Pet, Reporter
from .test_query import add_test_data, to_std_dicts
from alchql.middlewares import LoaderMiddleware
from alchql.types import SQLAlchemyObjectType


@pytest.mark.asyncio
async def test_query_pet_kinds(session):
    await add_test_data(session)

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet

    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter

        async def resolve_pets(self, _info):
            s = _info.context.session
            q = sa.select(Pet).where(Pet.reporter_id == self.id)
            _result = await s.execute(q)

            return _result.scalars().all()

    class Query(graphene.ObjectType):
        reporter = graphene.Field(ReporterType)
        reporters = graphene.List(ReporterType)
        pets = graphene.List(
            PetType, kind=graphene.Argument(PetType.enum_for_field("pet_kind"))
        )

        async def resolve_reporter(self, _info):
            s = _info.context.session
            _result = await s.execute(sa.select(Reporter))
            return _result.scalars().first()

        async def resolve_reporters(self, _info):
            s = _info.context.session
            _result = await s.execute(sa.select(Reporter))
            return _result.scalars().all()

        async def resolve_pets(self, _info, kind):
            s = _info.context.session
            q = sa.select(Pet)
            if kind:
                q = q.where(Pet.pet_kind == kind.value)

            _result = await s.execute(q)

            return _result.scalars().all()

    query = """
        query ReporterQuery {
          reporter {
            firstName
            lastName
            email
            favoritePetKind
            pets {
              name
              petKind
            }
          }
          reporters {
            firstName
            favoritePetKind
          }
          pets(kind: DOG) {
            name
            petKind
          }
        }
    """
    expected = {
        "reporter": {
            "firstName": "John",
            "lastName": "Doe",
            "email": None,
            "favoritePetKind": "CAT",
            "pets": [{"name": "Garfield", "petKind": "CAT"}],
        },
        "reporters": [
            {
                "firstName": "John",
                "favoritePetKind": "CAT",
            },
            {
                "firstName": "Jane",
                "favoritePetKind": "DOG",
            },
        ],
        "pets": [{"name": "Lassie", "petKind": "DOG"}],
    }
    schema = graphene.Schema(query=Query)
    result = await schema.execute_async(
        query,
        context=Context(session=session),
        middleware=[
            LoaderMiddleware([Reporter, Pet]),
        ],
    )
    assert not result.errors
    assert result.data == expected


@pytest.mark.asyncio
async def test_query_more_enums(session):
    await add_test_data(session)

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet

    class Query(graphene.ObjectType):
        pet = graphene.Field(PetType)

        async def resolve_pet(self, _info):
            _result = await session.execute(sa.select(Pet))
            return _result.scalars().first()

    query = """
        query PetQuery {
          pet {
            name,
            petKind
            hairKind
          }
        }
    """
    expected = {"pet": {"name": "Garfield", "petKind": "CAT", "hairKind": "SHORT"}}
    schema = graphene.Schema(query=Query)
    result = await schema.execute_async(query)
    assert not result.errors
    result = to_std_dicts(result.data)
    assert result == expected


@pytest.mark.asyncio
async def test_enum_as_argument(session):
    await add_test_data(session)

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet

    class Query(graphene.ObjectType):
        pet = graphene.Field(
            PetType, kind=graphene.Argument(PetType.enum_for_field("pet_kind"))
        )

        async def resolve_pet(self, info, kind=None):
            q = sa.select(Pet)
            if kind:
                q = q.where(Pet.pet_kind == kind.value)
            return (await session.execute(q)).scalars().first()

    query = """
        query PetQuery($kind: PetKind) {
          pet(kind: $kind) {
            name,
            petKind
            hairKind
          }
        }
    """

    schema = graphene.Schema(query=Query)
    result = await schema.execute_async(query, variables={"kind": "CAT"})
    assert not result.errors
    expected = {"pet": {"name": "Garfield", "petKind": "CAT", "hairKind": "SHORT"}}
    assert result.data == expected
    result = await schema.execute_async(query, variables={"kind": "DOG"})
    assert not result.errors
    expected = {"pet": {"name": "Lassie", "petKind": "DOG", "hairKind": "LONG"}}
    result = to_std_dicts(result.data)
    assert result == expected


@pytest.mark.asyncio
async def test_py_enum_as_argument(session):
    await add_test_data(session)

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet

    class Query(graphene.ObjectType):
        pet = graphene.Field(
            PetType,
            kind=graphene.Argument(PetType._meta.fields["hair_kind"].type.of_type),
        )

        async def resolve_pet(self, _info, kind=None):
            query = sa.select(Pet)
            if kind:
                # enum arguments are expected to be strings, not PyEnums
                query = query.where(Pet.hair_kind == HairKind(kind))
            return (await session.execute(query)).scalars().first()

    query = """
        query PetQuery($kind: HairKind) {
          pet(kind: $kind) {
            name,
            petKind
            hairKind
          }
        }
    """

    schema = graphene.Schema(query=Query)
    result = await schema.execute_async(query, variables={"kind": "SHORT"})
    assert not result.errors
    expected = {"pet": {"name": "Garfield", "petKind": "CAT", "hairKind": "SHORT"}}
    assert result.data == expected
    result = await schema.execute_async(query, variables={"kind": "LONG"})
    assert not result.errors
    expected = {"pet": {"name": "Lassie", "petKind": "DOG", "hairKind": "LONG"}}
    result = to_std_dicts(result.data)
    assert result == expected



import json

import graphene
import pytest
import sqlalchemy as sa
from graphene import Context
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine

from alchql.app import SessionQLApp
from alchql.fields import SQLAlchemyConnectionField
from alchql.middlewares import LoaderMiddleware
from alchql.node import AsyncNode
from alchql.sql_mutation import (
    SQLAlchemyCreateMutation,
)
from alchql.types import SQLAlchemyObjectType
from tests import models as m
from .models import Base, HairKind


async def get_all_pets(session, schema):
    result = await schema.execute_async(
        """
            query {
                allPets {
                    edges {
                        node {
                            id
                            name
                        }
                    }
                }
            }
        """,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([m.Pet]),
        ],
    )
    assert not result.errors

    all_pets = [i["node"] for i in result.data["allPets"]["edges"]]

    return all_pets


async def add_test_data(session):
    await session.execute(sa.delete(m.Pet))
    reporter_id = (
        await session.execute(
            sa.insert(m.Reporter).values(
                {
                    m.Reporter.first_name: "John",
                    m.Reporter.last_name: "Doe",
                    m.Reporter.favorite_pet_kind: "cat",
                }
            )
        )
    ).inserted_primary_key[0]

    await session.execute(
        sa.insert(m.Pet).values(
            {
                m.Pet.name: "Garfield",
                m.Pet.pet_kind: "cat",
                m.Pet.hair_kind: HairKind.SHORT,
                m.Pet.reporter_id: reporter_id,
            }
        )
    )

    await session.execute(
        sa.insert(m.Article).values(
            {
                m.Article.headline: "Hi!",
                m.Article.reporter_id: reporter_id,
            }
        )
    )

    reporter_id = (
        await session.execute(
            sa.insert(m.Reporter).values(
                {
                    m.Reporter.first_name: "Jane",
                    m.Reporter.last_name: "Roe",
                    m.Reporter.favorite_pet_kind: "dog",
                }
            )
        )
    ).inserted_primary_key[0]

    await session.execute(
        sa.insert(m.Pet).values(
            {
                m.Pet.name: "Lassie",
                m.Pet.pet_kind: "dog",
                m.Pet.hair_kind: HairKind.LONG,
                m.Pet.reporter_id: reporter_id,
            }
        )
    )

    await session.execute(
        sa.insert(m.Editor).values(
            {
                m.Editor.name: "Jack",
            }
        )
    )


@pytest.mark.asyncio
async def test_session_rollback():
    engine = create_async_engine(
        "sqlite+aiosqlite:///:memory:",
        convert_unicode=True,
        echo=False,
    )
    async with engine.begin() as con:
        await con.run_sync(Base.metadata.create_all)

    async with AsyncSession(engine) as session, session.begin():
        await add_test_data(session)

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = m.Pet
            interfaces = (AsyncNode,)

    class MutationInsertPet(SQLAlchemyCreateMutation):
        class Meta:
            model = m.Pet
            output = PetType

        @classmethod
        async def mutate(cls, root, info, value: dict):
            await super().mutate(root, info, value)
            raise Exception("Test error")

    class Query(graphene.ObjectType):
        node = AsyncNode.Field()
        all_pets = SQLAlchemyConnectionField(PetType.connection)

    class Mutation(graphene.ObjectType):
        insert_pet = MutationInsertPet.Field()

    schema = graphene.Schema(
        query=Query,
        mutation=Mutation,
    )

    async with AsyncSession(engine) as session:
        all_pets = await get_all_pets(session, schema)

    start_names = {i["name"] for i in all_pets}

    assert start_names == {
        "Garfield",
        "Lassie",
    }

    query = """
        mutation UpdatePet($value: MutationInsertPetInputType!) {
            insertPet(value: $value) {
                name
            }
        }
    """
    variables = {
        "value": {
            "name": "Odin",
            "petKind": "CAT",
            "hairKind": "SHORT",
        },
    }

    async def receive():
        return {
            "type": "http.request",
            "body": b'{"query": "%s", "variables": %s}'
            % (query.replace("\n", "").encode(), json.dumps(variables).encode()),
        }

    result = {}

    async def send(data):
        nonlocal result
        if data["type"] == "http.response.body":
            result = json.loads(data["body"])

    app = SessionQLApp(
        schema=schema,
        engine=engine,
    )

    await app(
        scope={
            "type": "http",
            "method": "POST",
            "headers": [(b"content-type", b"application/json")],
        },
        receive=receive,
        send=send,
    )

    assert result["errors"][0]["message"] == "Test error"

    all_pets = await get_all_pets(session, schema)
    assert {i["name"] for i in all_pets} == start_names


@pytest.mark.asyncio
@pytest.mark.skip
async def test_session_readonly(raise_graphql):
    engine = create_async_engine(
        "sqlite+aiosqlite:///:memory:",
        # "postgresql+asyncpg://postgres:postgres@127.0.0.1:5432/test",
        convert_unicode=True,
        echo=False,
    )
    async with engine.begin() as con:
        await con.run_sync(Base.metadata.drop_all)
    async with engine.begin() as con:
        await con.run_sync(Base.metadata.create_all)

    async with AsyncSession(engine) as session, session.begin():
        await add_test_data(session)

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = m.Pet
            interfaces = (AsyncNode,)

    class Query(graphene.ObjectType):
        node = AsyncNode.Field()
        all_pets = SQLAlchemyConnectionField(PetType.connection)

        async def resolve_all_pets(self, _info, *args, **kwargs):
            s = _info.context.session

            await s.execute(
                sa.insert(m.Pet).values(
                    {
                        m.Pet.name: "Test",
                        m.Pet.pet_kind: "cat",
                        m.Pet.hair_kind: "long",
                    }
                )
            )

            _result = (await s.execute(sa.select(m.Pet))).scalars().fetchall()
            return _result

    schema = graphene.Schema(query=Query)

    async with AsyncSession(engine) as session:
        all_pets = (await session.execute(sa.select(m.Pet.name))).scalars()

    start_names = {i for i in all_pets}

    assert start_names == {
        "Garfield",
        "Lassie",
    }

    query = """
        query {
            allPets {
                edges {
                    node { 
                        name 
                    }
                }
            }
        }
    """

    async def receive():
        return {
            "type": "http.request",
            "body": b'{"query": "%s"}' % query.replace("\n", "").encode(),
        }

    result = {}

    async def send(data):
        nonlocal result
        if data["type"] == "http.response.body":
            result = json.loads(data["body"])

    app = SessionQLApp(
        schema=schema,
        engine=engine,
        middleware=[
            LoaderMiddleware([m.Pet]),
        ],
    )

    await app(
        scope={
            "type": "http",
            "method": "POST",
            "headers": [(b"content-type", b"application/json")],
        },
        receive=receive,
        send=send,
    )

    assert not result.get("errors"), result["errors"][0]

    async with AsyncSession(engine) as session:
        all_pets = (await session.execute(sa.select(m.Pet.name))).scalars()

    assert {i for i in all_pets} == start_names



import graphene
import pytest
import sqlalchemy as sa
from graphene import Context

from .models import Article, HairKind, Pet, Reporter
from alchql.fields import BatchSQLAlchemyConnectionField
from alchql.middlewares import LoaderMiddleware
from alchql.node import AsyncNode
from alchql.types import SQLAlchemyObjectType


def get_schema():
    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter
            interfaces = (AsyncNode,)
            connection_field_factory = BatchSQLAlchemyConnectionField.from_relationship

    class ArticleType(SQLAlchemyObjectType):
        class Meta:
            model = Article
            interfaces = (AsyncNode,)
            connection_field_factory = BatchSQLAlchemyConnectionField.from_relationship

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            interfaces = (AsyncNode,)
            connection_field_factory = BatchSQLAlchemyConnectionField.from_relationship

    class Query(graphene.ObjectType):
        articles = graphene.Field(graphene.List(ArticleType))
        reporters = graphene.Field(graphene.List(ReporterType))

        async def resolve_articles(self, info):
            session = info.context.session
            result = await session.execute(sa.select(Article))
            return result.scalars()

        async def resolve_reporters(self, info):
            session = info.context.session
            result = await session.execute(sa.select(Reporter))
            return result.scalars().all()

    return graphene.Schema(query=Query)


async def benchmark_query(session, benchmark, query):
    schema = get_schema()

    @benchmark
    async def execute_query():
        result = await schema.execute_async(
            query,
            context_value=Context(session=session),
            middleware=[
                LoaderMiddleware([Article, Reporter, Pet]),
            ],
        )
        assert not result.errors


@pytest.mark.asyncio
async def test_one_to_one(session, benchmark):
    reporter_1 = Reporter(
        first_name="Reporter_1",
    )
    session.add(reporter_1)
    reporter_2 = Reporter(
        first_name="Reporter_2",
    )
    session.add(reporter_2)

    article_1 = Article(headline="Article_1")
    article_1.reporter = reporter_1
    session.add(article_1)

    article_2 = Article(headline="Article_2")
    article_2.reporter = reporter_2
    session.add(article_2)

    await session.commit()

    await benchmark_query(
        session,
        benchmark,
        """
      query {
        reporters {
          firstName
          favoriteArticle {
            headline
          }
        }
      }
    """,
    )


@pytest.mark.asyncio
async def test_many_to_one(session, benchmark):
    reporter_1 = Reporter(
        first_name="Reporter_1",
    )
    session.add(reporter_1)
    reporter_2 = Reporter(
        first_name="Reporter_2",
    )
    session.add(reporter_2)

    article_1 = Article(headline="Article_1")
    article_1.reporter = reporter_1
    session.add(article_1)

    article_2 = Article(headline="Article_2")
    article_2.reporter = reporter_2
    session.add(article_2)

    await session.commit()

    await benchmark_query(
        session,
        benchmark,
        """
      query {
        articles {
          headline
          reporter {
            firstName
          }
        }
      }
    """,
    )


@pytest.mark.asyncio
async def test_one_to_many(session, benchmark):
    reporter_1 = Reporter(
        first_name="Reporter_1",
    )
    session.add(reporter_1)
    reporter_2 = Reporter(
        first_name="Reporter_2",
    )
    session.add(reporter_2)

    article_1 = Article(headline="Article_1")
    article_1.reporter = reporter_1
    session.add(article_1)

    article_2 = Article(headline="Article_2")
    article_2.reporter = reporter_1
    session.add(article_2)

    article_3 = Article(headline="Article_3")
    article_3.reporter = reporter_2
    session.add(article_3)

    article_4 = Article(headline="Article_4")
    article_4.reporter = reporter_2
    session.add(article_4)

    await session.commit()

    await benchmark_query(
        session,
        benchmark,
        """
      query {
        reporters {
          firstName
          articles(first: 2) {
            edges {
              node {
                headline
              }
            }
          }
        }
      }
    """,
    )


@pytest.mark.asyncio
async def test_many_to_many(session, benchmark):
    reporter_1 = Reporter(
        first_name="Reporter_1",
    )
    session.add(reporter_1)
    reporter_2 = Reporter(
        first_name="Reporter_2",
    )
    session.add(reporter_2)

    pet_1 = Pet(name="Pet_1", pet_kind="cat", hair_kind=HairKind.LONG)
    session.add(pet_1)

    pet_2 = Pet(name="Pet_2", pet_kind="cat", hair_kind=HairKind.LONG)
    session.add(pet_2)

    reporter_1.pets.append(pet_1)
    reporter_1.pets.append(pet_2)

    pet_3 = Pet(name="Pet_3", pet_kind="cat", hair_kind=HairKind.LONG)
    session.add(pet_3)

    pet_4 = Pet(name="Pet_4", pet_kind="cat", hair_kind=HairKind.LONG)
    session.add(pet_4)

    reporter_2.pets.append(pet_3)
    reporter_2.pets.append(pet_4)

    await session.commit()

    await benchmark_query(
        session,
        benchmark,
        """
      query {
        reporters {
          firstName
          pets(first: 2) {
            edges {
              node {
                name
              }
            }
          }
        }
      }
    """,
    )



import contextlib
import logging
from unittest.mock import patch

import graphene
import pytest
import sqlalchemy as sa
from graphene import Context
from sqlalchemy.ext.asyncio import AsyncSession

from alchql.middlewares import LoaderMiddleware
from alchql.node import AsyncNode
from alchql.types import SQLAlchemyObjectType
from .models import Article, association_table, HairKind, Pet, Reporter
from .utils import to_std_dicts


class MockLoggingHandler(logging.Handler):
    """Intercept and store log messages in a list."""

    def __init__(self, *args, **kwargs):
        self.messages = []
        logging.Handler.__init__(self, *args, **kwargs)

    def emit(self, record):
        self.messages.append(record.getMessage())


@contextlib.contextmanager
def mock_sqlalchemy_logging_handler():
    logging.basicConfig()
    sql_logger = logging.getLogger("sqlalchemy.engine")
    previous_level = sql_logger.level

    sql_logger.setLevel(logging.INFO)
    mock_logging_handler = MockLoggingHandler()
    mock_logging_handler.setLevel(logging.INFO)
    sql_logger.addHandler(mock_logging_handler)

    yield mock_logging_handler

    sql_logger.setLevel(previous_level)


def get_schema():
    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter
            interfaces = (AsyncNode,)

    class ArticleType(SQLAlchemyObjectType):
        class Meta:
            model = Article
            interfaces = (AsyncNode,)

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            interfaces = (AsyncNode,)

    class Query(graphene.ObjectType):
        articles = graphene.Field(graphene.List(ArticleType))
        reporters = graphene.Field(graphene.List(ReporterType))
        pets = graphene.Field(graphene.List(PetType))

        async def resolve_articles(self, info):
            session = info.context.session
            result = await session.execute(sa.select(Article))
            return result.scalars().all()

        async def resolve_reporters(self, info):
            session = info.context.session
            result = await session.execute(sa.select(Reporter))
            return result.scalars().all()

        async def resolve_pets(self, info):
            session = info.context.session
            result = await session.execute(sa.select(Pet))
            return result.scalars().all()

    return graphene.Schema(query=Query)


@pytest.mark.asyncio
async def test_many_to_one(session, raise_graphql):
    await session.execute(sa.insert(Reporter).values({"first_name": "Reporter_1"}))
    await session.execute(sa.insert(Reporter).values({"first_name": "Reporter_2"}))

    await session.execute(
        sa.insert(Article).values(
            {
                Article.headline: "Article_1",
                Article.reporter_id: sa.select(Reporter.id).where(
                    Reporter.first_name == "Reporter_1"
                ),
            }
        )
    )
    await session.execute(
        sa.insert(Article).values(
            {
                Article.headline: "Article_2",
                Article.reporter_id: sa.select(Reporter.id).where(
                    Reporter.first_name == "Reporter_2"
                ),
            }
        )
    )

    schema = get_schema()

    with patch.object(AsyncSession, "execute", wraps=session.execute) as execute:
        result = await schema.execute_async(
            """
              query {
                articles {
                  headline
                  reporter {
                    firstName
                    articles {
                      edges {
                        node {
                          headline
                        }
                      }
                    }
                  }
                }
              }
            """,
            context_value=Context(session=session),
            middleware=[
                LoaderMiddleware([Article, Reporter]),
            ],
        )

    assert execute.call_count == 3

    assert not result.errors
    result = to_std_dicts(result.data)
    assert result == {
        "articles": [
            {
                "headline": "Article_1",
                "reporter": {
                    "articles": {"edges": [{"node": {"headline": "Article_1"}}]},
                    "firstName": "Reporter_1",
                },
            },
            {
                "headline": "Article_2",
                "reporter": {
                    "articles": {"edges": [{"node": {"headline": "Article_2"}}]},
                    "firstName": "Reporter_2",
                },
            },
        ]
    }


@pytest.mark.asyncio
async def test_one_to_one(session):
    await session.execute(
        sa.insert(Reporter),
        [
            {Reporter.first_name.key: "Reporter_1"},
            {Reporter.first_name.key: "Reporter_2"},
        ],
    )

    await session.execute(
        sa.insert(Article).values(
            {
                Article.headline: "Article_1",
                Article.reporter_id: sa.select(Reporter.id).where(
                    Reporter.first_name == "Reporter_1"
                ),
            }
        )
    )
    await session.execute(
        sa.insert(Article).values(
            {
                Article.headline: "Article_2",
                Article.reporter_id: sa.select(Reporter.id).where(
                    Reporter.first_name == "Reporter_2"
                ),
            }
        )
    )

    schema = get_schema()

    with patch.object(AsyncSession, "execute", wraps=session.execute) as execute:
        result = await schema.execute_async(
            """
              query {
                reporters {
                  firstName
                  favoriteArticle {
                    headline
                  }
                }
              }
            """,
            context_value=Context(session=session),
            middleware=[
                LoaderMiddleware([Article, Reporter]),
            ],
        )

    assert not result.errors, result.errors
    assert execute.call_count == 2

    assert not result.errors
    result = to_std_dicts(result.data)
    assert result == {
        "reporters": [
            {
                "firstName": "Reporter_1",
                "favoriteArticle": {
                    "headline": "Article_1",
                },
            },
            {
                "firstName": "Reporter_2",
                "favoriteArticle": {
                    "headline": "Article_2",
                },
            },
        ],
    }


@pytest.mark.asyncio
async def test_one_to_many(session, raise_graphql):
    await session.execute(
        sa.insert(Reporter),
        [
            {Reporter.first_name.key: "Reporter_1"},
            {Reporter.first_name.key: "Reporter_2"},
        ],
    )

    await session.execute(
        sa.insert(Article).values(
            {
                Article.headline: "Article_1",
                Article.reporter_id: sa.select(Reporter.id).where(
                    Reporter.first_name == "Reporter_1"
                ),
            }
        )
    )
    await session.execute(
        sa.insert(Article).values(
            {
                Article.headline: "Article_2",
                Article.reporter_id: sa.select(Reporter.id).where(
                    Reporter.first_name == "Reporter_1"
                ),
            }
        )
    )

    await session.execute(
        sa.insert(Article).values(
            {
                Article.headline: "Article_3",
                Article.reporter_id: sa.select(Reporter.id).where(
                    Reporter.first_name == "Reporter_2"
                ),
            }
        )
    )

    await session.execute(
        sa.insert(Article).values(
            {
                Article.headline: "Article_4",
                Article.reporter_id: sa.select(Reporter.id).where(
                    Reporter.first_name == "Reporter_2"
                ),
            }
        )
    )

    schema = get_schema()

    with patch.object(AsyncSession, "execute", wraps=session.execute) as execute:
        result = await schema.execute_async(
            """
            query {
                reporters {
                    firstName
                    articles(first: 2) {
                        edges {
                            node {
                                headline
                            }
                        }
                    }
                }
            }
            """,
            context_value=Context(session=session),
            middleware=[
                LoaderMiddleware([Article, Reporter]),
            ],
        )

    assert not result.errors, result.errors
    assert execute.call_count == 2

    assert not result.errors
    result = to_std_dicts(result.data)
    assert result == {
        "reporters": [
            {
                "firstName": "Reporter_1",
                "articles": {
                    "edges": [
                        {
                            "node": {
                                "headline": "Article_1",
                            },
                        },
                        {
                            "node": {
                                "headline": "Article_2",
                            },
                        },
                    ],
                },
            },
            {
                "firstName": "Reporter_2",
                "articles": {
                    "edges": [
                        {
                            "node": {
                                "headline": "Article_3",
                            },
                        },
                        {
                            "node": {
                                "headline": "Article_4",
                            },
                        },
                    ],
                },
            },
        ],
    }


@pytest.mark.asyncio
async def test_one_to_many_sorted(session, raise_graphql):
    await session.execute(
        sa.insert(Reporter),
        [
            {Reporter.first_name.key: "Reporter_1"},
            {Reporter.first_name.key: "Reporter_2"},
            {Reporter.first_name.key: "Reporter_3"},
        ],
    )

    reporter_1_id = (
        await session.execute(
            sa.select(Reporter.id).where(Reporter.first_name == "Reporter_1")
        )
    ).scalar()
    reporter_2_id = (
        await session.execute(
            sa.select(Reporter.id).where(Reporter.first_name == "Reporter_2")
        )
    ).scalar()

    await session.execute(
        sa.insert(Article).values(
            [
                {Article.headline: "Article_1", Article.reporter_id: reporter_1_id},
                {Article.headline: "Article_2", Article.reporter_id: reporter_1_id},
                {Article.headline: "Article_3", Article.reporter_id: reporter_2_id},
                {Article.headline: "Article_4", Article.reporter_id: reporter_2_id},
            ]
        )
    )

    schema = get_schema()

    # Passing sort inside the query
    with patch.object(AsyncSession, "execute", wraps=session.execute) as execute:
        result = await schema.execute_async(
            """
            query {
                reporters {
                    firstName
                    articles(first: 2, sort: HEADLINE_DESC) {
                        edges {
                            node {
                                headline
                            }
                        }
                    }
                }
            }
            """,
            context_value=Context(session=session),
            middleware=[
                LoaderMiddleware([Article, Reporter]),
            ],
        )

    assert not result.errors, result.errors
    assert execute.call_count == 2

    result = to_std_dicts(result.data)
    expected_result = {
        "reporters": [
            {
                "firstName": "Reporter_1",
                "articles": {
                    "edges": [
                        {"node": {"headline": "Article_2"}},
                        {"node": {"headline": "Article_1"}},
                    ],
                },
            },
            {
                "firstName": "Reporter_2",
                "articles": {
                    "edges": [
                        {"node": {"headline": "Article_4"}},
                        {"node": {"headline": "Article_3"}},
                    ],
                },
            },
            {
                "firstName": "Reporter_3",
                "articles": {
                    "edges": [],
                },
            },
        ],
    }
    assert result == expected_result

    # Passing sort in variables
    with patch.object(AsyncSession, "execute", wraps=session.execute) as execute:
        result = await schema.execute_async(
            """
            query($sort: [ArticleTypeSortEnum]) {
                reporters {
                    firstName
                    articles(first: 2, sort: $sort) {
                        edges {
                            node {
                                headline
                            }
                        }
                    }
                }
            }
            """,
            context_value=Context(session=session),
            middleware=[
                LoaderMiddleware([Article, Reporter]),
            ],
            variables={"sort": "HEADLINE_DESC"},
        )

    assert not result.errors, result.errors
    assert execute.call_count == 2

    result = to_std_dicts(result.data)
    assert expected_result == result


@pytest.mark.asyncio
async def test_many_to_many(session):
    await session.execute(
        sa.insert(Reporter),
        [
            {Reporter.first_name.key: "Reporter_1"},
            {Reporter.first_name.key: "Reporter_2"},
        ],
    )
    reporters = dict(
        (await session.execute(sa.select(Reporter.first_name, Reporter.id))).fetchall()
    )

    await session.execute(
        sa.insert(Pet),
        [
            {
                Pet.name.key: "Pet_1",
                Pet.pet_kind.key: "cat",
                Pet.hair_kind.key: HairKind.LONG,
            },
            {
                Pet.name.key: "Pet_2",
                Pet.pet_kind.key: "cat",
                Pet.hair_kind.key: HairKind.LONG,
            },
            {
                Pet.name.key: "Pet_3",
                Pet.pet_kind.key: "cat",
                Pet.hair_kind.key: HairKind.LONG,
            },
            {
                Pet.name.key: "Pet_4",
                Pet.pet_kind.key: "cat",
                Pet.hair_kind.key: HairKind.LONG,
            },
        ],
    )
    pets = dict((await session.execute(sa.select(Pet.name, Pet.id))).fetchall())

    await session.execute(
        sa.insert(association_table),
        [
            {
                association_table.c.pet_id.key: pets["Pet_1"],
                association_table.c.reporter_id.key: reporters["Reporter_1"],
            },
            {
                association_table.c.pet_id.key: pets["Pet_2"],
                association_table.c.reporter_id.key: reporters["Reporter_1"],
            },
            {
                association_table.c.pet_id.key: pets["Pet_3"],
                association_table.c.reporter_id.key: reporters["Reporter_2"],
            },
            {
                association_table.c.pet_id.key: pets["Pet_4"],
                association_table.c.reporter_id.key: reporters["Reporter_2"],
            },
        ],
    )

    schema = get_schema()

    with patch.object(AsyncSession, "execute", wraps=session.execute) as execute:
        result = await schema.execute_async(
            """
              query {
                reporters {
                  firstName
                  pets(first: 2) {
                    edges {
                      node {
                        name
                      }
                    }
                  }
                }
              }
            """,
            context_value=Context(session=session),
            middleware=[
                LoaderMiddleware([Article, Reporter, Pet]),
            ],
        )

    assert execute.call_count == 2

    assert not result.errors, result.errors[0]
    result = to_std_dicts(result.data)
    assert result == {
        "reporters": [
            {
                "firstName": "Reporter_1",
                "pets": {
                    "edges": [
                        {"node": {"name": "Pet_1"}},
                        {"node": {"name": "Pet_2"}},
                    ],
                },
            },
            {
                "firstName": "Reporter_2",
                "pets": {
                    "edges": [
                        {"node": {"name": "Pet_3"}},
                        {"node": {"name": "Pet_4"}},
                    ],
                },
            },
        ],
    }

    with patch.object(AsyncSession, "execute", wraps=session.execute) as execute:
        result = await schema.execute_async(
            """
              query {
                pets {
                  name
                  reporters(first: 2) {
                    edges {
                      node {
                        firstName
                      }
                    }
                  }
                }
              }
            """,
            context_value=Context(session=session),
            middleware=[
                LoaderMiddleware([Article, Reporter, Pet]),
            ],
        )

    assert not result.errors, result.errors[0]
    assert execute.call_count == 2

    result = to_std_dicts(result.data)
    assert result == {
        "pets": [
            {
                "name": "Pet_1",
                "reporters": {"edges": [{"node": {"firstName": "Reporter_1"}}]},
            },
            {
                "name": "Pet_2",
                "reporters": {"edges": [{"node": {"firstName": "Reporter_1"}}]},
            },
            {
                "name": "Pet_3",
                "reporters": {"edges": [{"node": {"firstName": "Reporter_2"}}]},
            },
            {
                "name": "Pet_4",
                "reporters": {"edges": [{"node": {"firstName": "Reporter_2"}}]},
            },
        ]
    }


@pytest.mark.asyncio
async def test_many_to_many_sorted(session):
    await session.execute(
        sa.insert(Reporter),
        [
            {Reporter.first_name.key: "Reporter_1"},
            {Reporter.first_name.key: "Reporter_2"},
        ],
    )
    reporters = dict(
        (await session.execute(sa.select(Reporter.first_name, Reporter.id))).fetchall()
    )

    pet_reporter = {
        "Pet_1": "Reporter_1",
        "Pet_2": "Reporter_1",
        "Pet_3": "Reporter_2",
        "Pet_4": "Reporter_2",
    }

    await session.execute(
        sa.insert(Pet),
        [
            {
                Pet.name.key: pet_name,
                Pet.pet_kind.key: "cat",
                Pet.hair_kind.key: HairKind.LONG,
            }
            for pet_name in pet_reporter
        ],
    )
    pets = dict((await session.execute(sa.select(Pet.name, Pet.id))).fetchall())

    await session.execute(
        sa.insert(association_table),
        [
            {
                association_table.c.pet_id.key: pets[pet_name],
                association_table.c.reporter_id.key: reporters[reporter_name],
            }
            for pet_name, reporter_name in pet_reporter.items()
        ],
    )

    expected_result = {
        "reporters": [
            {
                "firstName": "Reporter_1",
                "pets": {
                    "edges": [
                        {"node": {"name": "Pet_2"}},
                        {"node": {"name": "Pet_1"}},
                    ]
                },
            },
            {
                "firstName": "Reporter_2",
                "pets": {
                    "edges": [
                        {"node": {"name": "Pet_4"}},
                        {"node": {"name": "Pet_3"}},
                    ]
                },
            },
        ],
    }
    schema = get_schema()

    # Passing sort inside the query
    with patch.object(AsyncSession, "execute", wraps=session.execute) as execute:
        result = await schema.execute_async(
            """
              query {
                reporters {
                  firstName
                  pets(first: 2, sort: NAME_DESC) {
                    edges {
                      node {
                        name
                      }
                    }
                  }
                }
              }
            """,
            context_value=Context(session=session),
            middleware=[
                LoaderMiddleware([Article, Reporter, Pet]),
            ],
        )

    assert not result.errors, result.errors[0]
    assert execute.call_count == 2

    result = to_std_dicts(result.data)
    assert result == expected_result

    # Passing sort in variables
    with patch.object(AsyncSession, "execute", wraps=session.execute) as execute:
        result = await schema.execute_async(
            """
              query($sort: [PetTypeSortEnum]) {
                reporters {
                  firstName
                  pets(first: 2, sort: $sort) {
                    edges {
                      node {
                        name
                      }
                    }
                  }
                }
              }
            """,
            context_value=Context(session=session),
            middleware=[
                LoaderMiddleware([Article, Reporter, Pet]),
            ],
            variables={"sort": "NAME_DESC"},
        )

    assert not result.errors, result.errors[0]
    assert execute.call_count == 2

    result = to_std_dicts(result.data)
    assert result == expected_result


@pytest.mark.asyncio
async def test_only_ids(session, raise_graphql):
    await session.execute(
        sa.insert(Reporter),
        [
            {Reporter.first_name.key: "Reporter_1"},
            {Reporter.first_name.key: "Reporter_2"},
            {Reporter.first_name.key: "Reporter_3"},
        ],
    )

    reporter_1_id = (
        await session.execute(
            sa.select(Reporter.id).where(Reporter.first_name == "Reporter_1")
        )
    ).scalar()
    reporter_2_id = (
        await session.execute(
            sa.select(Reporter.id).where(Reporter.first_name == "Reporter_2")
        )
    ).scalar()

    await session.execute(
        sa.insert(Article).values(
            [
                {Article.headline: "Article_1", Article.reporter_id: reporter_1_id},
                {Article.headline: "Article_2", Article.reporter_id: reporter_1_id},
                {Article.headline: "Article_3", Article.reporter_id: reporter_2_id},
                {Article.headline: "Article_4", Article.reporter_id: reporter_2_id},
            ]
        )
    )

    schema = get_schema()

    old_exec = AsyncSession.execute

    async def execute_mock(self, command):
        return await old_exec(self, command)

    with patch.object(AsyncSession, "execute", execute_mock) as execute:
        result = await schema.execute_async(
            """
            query {
                reporters {
                    articles(first: 2, sort: HEADLINE_DESC) {
                        edges {
                            node {
                                id
                            }
                        }
                    }
                }
            }
            """,
            context_value=Context(session=session),
            middleware=[
                LoaderMiddleware([Reporter, Article]),
            ],
        )

    assert not result.errors

    assert not result.errors
    result = to_std_dicts(result.data)
    assert result == {
        "reporters": [
            {
                "articles": {
                    "edges": [
                        {"node": {"id": "QXJ0aWNsZVR5cGU6Mg=="}},
                        {"node": {"id": "QXJ0aWNsZVR5cGU6MQ=="}},
                    ]
                }
            },
            {
                "articles": {
                    "edges": [
                        {"node": {"id": "QXJ0aWNsZVR5cGU6NA=="}},
                        {"node": {"id": "QXJ0aWNsZVR5cGU6Mw=="}},
                    ]
                }
            },
            {"articles": {"edges": []}},
        ]
    }



import re
from contextlib import asynccontextmanager
from inspect import isawaitable
from typing import Any, Awaitable, Callable, Dict, List, Optional, Type, Union

from graphene import Context
from graphql import ExecutionResult, graphql
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, AsyncSessionTransaction
from starlette.background import BackgroundTasks
from starlette.requests import HTTPConnection, Request
from starlette.responses import HTMLResponse, JSONResponse, Response
from starlette_graphene3 import (
    GraphQLApp,
    _get_operation_from_request,
)

from .extensions import Extension, ExtensionManager

DEFAULT_GET = object()

QUERY_REGEX = re.compile(r"^\s*?(query)", flags=re.M)


class SessionQLApp(GraphQLApp):
    def __init__(
        self,
        engine: AsyncEngine,
        context_value: Callable = Context,
        on_get: Optional[
            Callable[[Request], Union[Response, Awaitable[Response]]]
        ] = DEFAULT_GET,
        extensions: List[Type[Extension]] = (),
        raise_exceptions: List[Type[Exception]] = (),
        *args,
        **kwargs,
    ):
        self.engine = engine
        if on_get == DEFAULT_GET:
            on_get = lambda request: HTMLResponse(
                f"""
                <html>
                <head>
                    <script 
                        type="application/javascript" 
                        src="https://embeddable-sandbox.cdn.apollographql.com/_latest/embeddable-sandbox.umd.production.min.js"
                    ></script>
                </head>
                <body style="margin:0">
                    <div style="width: 100%; height: 100%;" id='embedded-sandbox'></div>
                    <script>
                      new window.EmbeddedSandbox({{
                        target: '#embedded-sandbox',
                        initialEndpoint: '{request.url}'
                      }});
                    </script>
                </body>
                </html>
                """
            )

        self.extensions = extensions or ()
        self.raise_exceptions = tuple(raise_exceptions) or ()
        super().__init__(context_value=context_value, on_get=on_get, *args, **kwargs)

    async def _handle_http_request(self, request: Request) -> JSONResponse:
        try:
            operations = await _get_operation_from_request(request)
        except ValueError as e:
            return JSONResponse({"errors": [e.args[0]]}, status_code=400)

        if isinstance(operations, list):
            return JSONResponse(
                {"errors": ["This server does not support batching"]}, status_code=400
            )
        else:
            operation = operations

        query = operation["query"]
        variable_values = operation.get("variables")
        operation_name = operation.get("operationName")

        is_ro_operation = QUERY_REGEX.search(query) is not None

        async with self._get_context_value(
            request
        ) as context_value, self._get_transaction(is_ro_operation) as transaction:
            context_value.session = transaction.session

            middleware = self.middleware or ()
            extension_manager = ExtensionManager(self.extensions, context=context_value)

            with extension_manager.request():
                result: ExecutionResult = await graphql(
                    self.schema.graphql_schema,
                    source=query,
                    context_value=context_value,
                    root_value=self.root_value,
                    middleware=(*middleware, *extension_manager.extensions),
                    variable_values=variable_values,
                    operation_name=operation_name,
                    execution_context_class=self.execution_context_class,
                )

            if result.errors:
                await transaction.rollback()

            extension_results = extension_manager.format()
            if extension_results:
                result.extensions = extension_results

            background = getattr(context_value, "background", None)

        response: Dict[str, Any] = {"data": result.data}
        if result.errors:
            for error in result.errors:
                if error.original_error:
                    self.logger.error(
                        "An exception occurred in resolvers",
                        exc_info=error.original_error,
                    )
                    if isinstance(error.original_error, self.raise_exceptions):
                        raise error.original_error
            response["errors"] = [
                self.error_formatter(error) for error in result.errors
            ]
        if result.extensions:
            response["extensions"] = result.extensions

        return JSONResponse(
            response,
            status_code=200,
            background=background,
        )

    @asynccontextmanager
    async def _get_context_value(self, request: HTTPConnection) -> Context:
        if callable(self.context_value):
            context = self.context_value(
                request=request,
                background=BackgroundTasks(),
            )
            if isawaitable(context):
                context = await context
            yield context
        else:
            yield self.context_value or Context(
                request=request,
                background=BackgroundTasks(),
            )

    @asynccontextmanager
    async def _get_transaction(self, is_ro_operation: bool) -> AsyncSessionTransaction:
        async with AsyncSession(self.engine) as session, session.begin() as transaction:
            execution_options = {}

            if is_ro_operation:
                execution_options["isolation_level"] = "AUTOCOMMIT"
                # does not work with AUTOCOMMIT
                # if getattr(session.bind, "name", "") == "postgresql":
                #     execution_options["postgresql_readonly"] = True
                #     execution_options["postgresql_deferrable"] = True

            if execution_options:
                await session.connection(execution_options=execution_options)

            yield transaction



from alchql.gql_fields import camel_to_snake


def test_camel_to_snake():
    assert camel_to_snake("artistId_Eq") == "artist_id__eq"
    assert camel_to_snake("sort") == "sort"
    assert camel_to_snake("priceMomentum12mo") == "price_momentum_12mo"
    assert camel_to_snake("s3PresignedUrl") == "s3_presigned_url"
    assert camel_to_snake("S3PresignedUrl") == "s3_presigned_url"
    assert camel_to_snake("thumbnailS3PresignedUrl") == "thumbnail_s3_presigned_url"



import pytest
from graphene import NonNull, ObjectType
from graphene.relay import Connection, Node

from .models import Editor as EditorModel, Pet as PetModel
from alchql.fields import SQLAlchemyConnectionField
from alchql.types import SQLAlchemyObjectType


class Pet(SQLAlchemyObjectType):
    class Meta:
        model = PetModel
        interfaces = (Node,)


class Editor(SQLAlchemyObjectType):
    class Meta:
        model = EditorModel


##
# SQLAlchemyConnectionField
##


def test_nonnull_sqlalachemy_connection():
    field = SQLAlchemyConnectionField(NonNull(Pet.connection))
    assert isinstance(field.type, NonNull)
    assert issubclass(field.type.of_type, Connection)
    assert field.type.of_type._meta.node is Pet


def test_required_sqlalachemy_connection():
    field = SQLAlchemyConnectionField(Pet.connection, required=True)
    assert isinstance(field.type, NonNull)
    assert issubclass(field.type.of_type, Connection)
    assert field.type.of_type._meta.node is Pet


def test_type_assert_sqlalchemy_object_type():
    with pytest.raises(AssertionError, match="only accepts SQLAlchemyObjectType"):
        SQLAlchemyConnectionField(ObjectType).type


def test_type_assert_object_has_connection():
    with pytest.raises(AssertionError, match="doesn't have a connection"):
        SQLAlchemyConnectionField(Editor).type


##
# UnsortedSQLAlchemyConnectionField
##


def test_sort_added_by_default():
    field = SQLAlchemyConnectionField(Pet.connection)
    assert "sort" in field.args
    assert field.args["sort"] == Pet.sort_argument()


def test_sort_can_be_removed():
    field = SQLAlchemyConnectionField(Pet.connection, sort=None)
    assert "sort" not in field.args


def test_custom_sort():
    field = SQLAlchemyConnectionField(Pet.connection, sort=Editor.sort_argument())
    assert field.args["sort"] == Editor.sort_argument()


def test_sort_init_raises():
    with pytest.raises(TypeError, match="Cannot create sort"):
        SQLAlchemyConnectionField(Connection)



from graphene import ObjectType

from .models import ReflectedEditor
from alchql.registry import Registry
from alchql.types import SQLAlchemyObjectType

registry = Registry()


class Reflected(SQLAlchemyObjectType):
    class Meta:
        model = ReflectedEditor
        registry = registry


def test_objecttype_registered():
    assert issubclass(Reflected, ObjectType)
    assert Reflected._meta.model == ReflectedEditor
    assert list(Reflected._meta.fields.keys()) == ["editor_id", "name"]



import pytest
import sqlalchemy as sa
from graphene import (
    Context,
    ObjectType,
    Schema,
)
from sqlalchemy.ext.asyncio import AsyncSession

from alchql.consts import OP_EQ
from alchql.fields import (
    BatchSQLAlchemyConnectionField,
    FilterConnectionField,
)
from alchql.middlewares import LoaderMiddleware
from alchql.node import AsyncNode
from alchql.types import (
    SQLAlchemyObjectType,
)
from .models import Reporter


async def add_reporter(session: AsyncSession):
    result = await session.execute(
        sa.insert(Reporter).values(
            {
                "first_name": "first_name",
                "last_name": "last_name",
                "email": "email",
                "favorite_pet_kind": "cat",
            }
        )
    )
    return result.lastrowid


@pytest.mark.asyncio
async def test_filter1(session):
    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter
            connection_field_factory = BatchSQLAlchemyConnectionField.from_relationship
            interfaces = (AsyncNode,)
            filter_fields = {
                Reporter.first_name: [OP_EQ],
            }

    class Query(ObjectType):
        reporter = FilterConnectionField(ReporterType, sort=None)

    await add_reporter(session)

    schema = Schema(query=Query, types=[ReporterType])
    result = await schema.execute_async(
        """
        query {
            reporter(firstName_Eq: "first_name") {
                edges{
                    node{
                        firstName
                    }
                }
            }
        }
        """,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Reporter]),
        ],
    )

    assert not result.errors
    assert result.data["reporter"]["edges"][0]["node"]["firstName"] == "first_name"


@pytest.mark.asyncio
async def test_filter2(session):
    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter
            connection_field_factory = BatchSQLAlchemyConnectionField.from_relationship
            interfaces = (AsyncNode,)
            filter_fields = {
                Reporter.__table__.c.first_name: [OP_EQ],
            }

    class Query(ObjectType):
        reporter = FilterConnectionField(ReporterType, sort=None)

    await add_reporter(session)

    schema = Schema(query=Query, types=[ReporterType])
    result = await schema.execute_async(
        """
        query {
            reporter(firstName_Eq: "first_name") {
                edges{
                    node{
                        firstName
                    }
                }
            }
        }
        """,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Reporter]),
        ],
    )

    assert not result.errors
    assert result.data["reporter"]["edges"][0]["node"]["firstName"] == "first_name"


@pytest.mark.asyncio
async def test_filter3(session):
    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter
            connection_field_factory = BatchSQLAlchemyConnectionField.from_relationship
            interfaces = (AsyncNode,)
            filter_fields = {
                Reporter.first_name.label("name_first"): [OP_EQ],
            }

    class Query(ObjectType):
        reporter = FilterConnectionField(ReporterType, sort=None)

    await add_reporter(session)

    schema = Schema(query=Query, types=[ReporterType])
    result = await schema.execute_async(
        """
        query {
            reporter(nameFirst_Eq: "first_name") {
                edges{
                    node{
                        firstName
                    }
                }
            }
        }
        """,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Reporter]),
        ],
    )

    assert not result.errors
    assert result.data["reporter"]["edges"][0]["node"]["firstName"] == "first_name"



from typing import Iterable, List, Type, Union

import graphene
import sqlalchemy as sa
from graphene.types.enum import EnumMeta
from graphql import StringValueNode
from sqlalchemy.orm import DeclarativeMeta

from .gql_id import ResolvedGlobalId
from .sqlalchemy_converter import convert_sqlalchemy_type


class ArgID(graphene.Scalar):
    @staticmethod
    def coerce_id(value):
        global_id = ResolvedGlobalId.decode(value)

        # registry = get_global_registry()
        # type_ = registry.get_type_for_model(table)
        # if type_ and type_.__name__ != global_id.type:
        #     raise Exception(
        #         f"Invalid GlobalID type: {global_id.type} != {type_.__name__}"
        #     )

        return global_id.id

    serialize = coerce_id
    parse_value = coerce_id

    @staticmethod
    def parse_literal(ast):
        if isinstance(ast, StringValueNode):
            return ast.value


def convert_sqlalchemy_type_mutation(column):
    field = convert_sqlalchemy_type(
        getattr(column, "type", None),
        column,
    )
    if field == graphene.Int and column.foreign_keys:
        field = ArgID()

    return field


def get_input_fields(
    model: Type[DeclarativeMeta],
    only_fields: Iterable[str] = (),
    exclude_fields: Iterable[str] = (),
    required_fields: Iterable[str] = (),
) -> dict:
    only_fields = set(only_fields)
    exclude_fields = set(exclude_fields)
    required_fields = set(required_fields)

    if only_fields and exclude_fields:
        raise ValueError(
            "The options 'only_fields' and 'exclude_fields' cannot be both set on the same type."
        )

    table = sa.inspect(model).persist_selectable
    fields = {}
    for name, column in dict(table.columns).items():
        if name not in required_fields:
            if only_fields and name not in only_fields:
                continue
            if exclude_fields and name in exclude_fields:
                continue

        field = convert_sqlalchemy_type_mutation(column)

        if callable(field):
            field = field()
        if isinstance(field, EnumMeta):
            field = field()

        if name in required_fields:
            getattr(field, "kwargs", {})["required"] = True

        fields[name] = field

    return fields


def get_input_type(name: str, input_fields: dict) -> Type:
    return type(name, (graphene.InputObjectType,), input_fields)



import pytest
import sqlalchemy as sa
from graphene import Argument, Context, Enum, List, ObjectType, Schema
from graphene.relay import Node

from alchql.fields import SQLAlchemyConnectionField
from alchql.middlewares import LoaderMiddleware
from alchql.types import SQLAlchemyObjectType
from alchql.utils import to_type_name
from .models import Base, HairKind, Pet
from .test_query import to_std_dicts


async def add_pets(session):
    q = sa.insert(Pet).values(
        [
            {"id": 1, "name": "Lassie", "pet_kind": "dog", "hair_kind": HairKind.LONG},
            {"id": 2, "name": "Barf", "pet_kind": "dog", "hair_kind": HairKind.LONG},
            {"id": 3, "name": "Alf", "pet_kind": "cat", "hair_kind": HairKind.LONG},
        ]
    )
    await session.execute(q)


def test_sort_enum():
    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet

    sort_enum = PetType.sort_enum()
    assert isinstance(sort_enum, type(Enum))
    assert sort_enum._meta.name == "PetTypeSortEnum"
    assert list(sort_enum._meta.enum.__members__) == [
        "ID_ASC",
        "ID_DESC",
        "NAME_ASC",
        "NAME_DESC",
        "PET_KIND_ASC",
        "PET_KIND_DESC",
        "HAIR_KIND_ASC",
        "HAIR_KIND_DESC",
        "REPORTER_ID_ASC",
        "REPORTER_ID_DESC",
    ]
    assert str(sort_enum.ID_ASC.value) == "pets.id ASC NULLS LAST"
    assert str(sort_enum.ID_DESC.value) == "pets.id DESC NULLS LAST"
    assert str(sort_enum.HAIR_KIND_ASC.value) == "pets.hair_kind ASC NULLS LAST"
    assert str(sort_enum.HAIR_KIND_DESC.value) == "pets.hair_kind DESC NULLS LAST"


def test_sort_enum_with_custom_name():
    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet

    sort_enum = PetType.sort_enum(name="CustomSortName")
    assert isinstance(sort_enum, type(Enum))
    assert sort_enum._meta.name == "CustomSortName"


def test_sort_enum_cache():
    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet

    sort_enum = PetType.sort_enum()
    sort_enum_2 = PetType.sort_enum()
    assert sort_enum_2 is sort_enum
    sort_enum_2 = PetType.sort_enum(name="PetTypeSortEnum")
    assert sort_enum_2 is sort_enum
    err_msg = "Sort enum for PetType has already been customized"
    with pytest.raises(ValueError, match=err_msg):
        PetType.sort_enum(name="CustomSortName")
    with pytest.raises(ValueError, match=err_msg):
        PetType.sort_enum(only_fields=["id"])
    with pytest.raises(ValueError, match=err_msg):
        PetType.sort_enum(only_indexed=True)
    with pytest.raises(ValueError, match=err_msg):
        PetType.sort_enum(get_symbol_name=lambda: "foo")


def test_sort_enum_with_excluded_field_in_object_type():
    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            exclude_fields = ["reporter_id"]

    sort_enum = PetType.sort_enum()
    assert list(sort_enum._meta.enum.__members__) == [
        "ID_ASC",
        "ID_DESC",
        "NAME_ASC",
        "NAME_DESC",
        "PET_KIND_ASC",
        "PET_KIND_DESC",
        "HAIR_KIND_ASC",
        "HAIR_KIND_DESC",
    ]


def test_sort_enum_only_fields():
    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet

    sort_enum = PetType.sort_enum(only_fields=["id", "name"])
    assert list(sort_enum._meta.enum.__members__) == [
        "ID_ASC",
        "ID_DESC",
        "NAME_ASC",
        "NAME_DESC",
    ]


def test_sort_argument():
    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet

    sort_arg = PetType.sort_argument()
    assert isinstance(sort_arg, Argument)

    assert isinstance(sort_arg.type, List)
    sort_enum = sort_arg.type._of_type
    assert isinstance(sort_enum, type(Enum))
    assert sort_enum._meta.name == "PetTypeSortEnum"
    assert list(sort_enum._meta.enum.__members__) == [
        "ID_ASC",
        "ID_DESC",
        "NAME_ASC",
        "NAME_DESC",
        "PET_KIND_ASC",
        "PET_KIND_DESC",
        "HAIR_KIND_ASC",
        "HAIR_KIND_DESC",
        "REPORTER_ID_ASC",
        "REPORTER_ID_DESC",
    ]
    assert str(sort_enum.ID_ASC.value) == "pets.id ASC NULLS LAST"
    assert str(sort_enum.ID_DESC.value) == "pets.id DESC NULLS LAST"
    assert str(sort_enum.HAIR_KIND_ASC.value) == "pets.hair_kind ASC NULLS LAST"
    assert str(sort_enum.HAIR_KIND_DESC.value) == "pets.hair_kind DESC NULLS LAST"

    assert list(map(str, sort_arg.default_value)) == ["pets.id ASC NULLS LAST"]
    assert str(sort_enum.ID_ASC.value) == "pets.id ASC NULLS LAST"


def test_sort_argument_with_excluded_fields_in_object_type():
    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            exclude_fields = ["hair_kind", "reporter_id"]

    sort_arg = PetType.sort_argument()
    sort_enum = sort_arg.type._of_type
    assert list(sort_enum._meta.enum.__members__) == [
        "ID_ASC",
        "ID_DESC",
        "NAME_ASC",
        "NAME_DESC",
        "PET_KIND_ASC",
        "PET_KIND_DESC",
    ]
    assert list(map(str, sort_arg.default_value)) == ["pets.id ASC NULLS LAST"]


def test_sort_argument_only_fields():
    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            only_fields = ["id", "pet_kind"]

    sort_arg = PetType.sort_argument()
    sort_enum = sort_arg.type._of_type
    assert list(sort_enum._meta.enum.__members__) == [
        "ID_ASC",
        "ID_DESC",
        "PET_KIND_ASC",
        "PET_KIND_DESC",
    ]
    assert list(map(str, sort_arg.default_value)) == ["pets.id ASC NULLS LAST"]


def test_sort_argument_for_multi_column_pk():
    class MultiPkTestModel(Base):
        __tablename__ = "multi_pk_test_table"
        foo = sa.Column(sa.Integer, primary_key=True)
        bar = sa.Column(sa.Integer, primary_key=True)

    class MultiPkTestType(SQLAlchemyObjectType):
        class Meta:
            model = MultiPkTestModel

    sort_arg = MultiPkTestType.sort_argument()
    assert list(map(str, sort_arg.default_value)) == [
        "multi_pk_test_table.foo ASC NULLS LAST",
        "multi_pk_test_table.bar ASC NULLS LAST",
    ]


def test_sort_argument_only_indexed():
    class IndexedTestModel(Base):
        __tablename__ = "indexed_test_table"
        id = sa.Column(sa.Integer, primary_key=True)
        foo = sa.Column(sa.Integer, index=False)
        bar = sa.Column(sa.Integer, index=True)

    class IndexedTestType(SQLAlchemyObjectType):
        class Meta:
            model = IndexedTestModel

    sort_arg = IndexedTestType.sort_argument(only_indexed=True)
    sort_enum = sort_arg.type._of_type
    assert list(sort_enum._meta.enum.__members__) == [
        "ID_ASC",
        "ID_DESC",
        "BAR_ASC",
        "BAR_DESC",
    ]
    assert list(map(str, sort_arg.default_value)) == [
        "indexed_test_table.id ASC NULLS LAST"
    ]


def test_sort_argument_with_custom_symbol_names():
    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet

    def get_symbol_name(column_name, sort_asc=True):
        return to_type_name(column_name) + ("Up" if sort_asc else "Down")

    sort_arg = PetType.sort_argument(get_symbol_name=get_symbol_name)
    sort_enum = sort_arg.type._of_type
    assert list(sort_enum._meta.enum.__members__) == [
        "IdUp",
        "IdDown",
        "NameUp",
        "NameDown",
        "PetKindUp",
        "PetKindDown",
        "HairKindUp",
        "HairKindDown",
        "ReporterIdUp",
        "ReporterIdDown",
    ]
    assert list(map(str, sort_arg.default_value)) == ["pets.id ASC NULLS LAST"]


@pytest.mark.asyncio
async def test_sort_query(session):
    await add_pets(session)

    class PetNode(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            interfaces = (Node,)

    class Query(ObjectType):
        defaultSort = SQLAlchemyConnectionField(PetNode.connection)
        nameSort = SQLAlchemyConnectionField(PetNode.connection)
        multipleSort = SQLAlchemyConnectionField(PetNode.connection)
        descSort = SQLAlchemyConnectionField(PetNode.connection)
        singleColumnSort = SQLAlchemyConnectionField(
            PetNode.connection, sort=Argument(PetNode.sort_enum())
        )
        noDefaultSort = SQLAlchemyConnectionField(
            PetNode.connection, sort=PetNode.sort_argument(has_default=False)
        )
        noSort = SQLAlchemyConnectionField(PetNode.connection, sort=None)

    query = """
        query sortTest {
            defaultSort {
                edges {
                    node {
                        name
                    }
                }
            }
            nameSort(sort: NAME_ASC) {
                edges {
                    node {
                        name
                    }
                }
            }
            multipleSort(sort: [PET_KIND_ASC, NAME_DESC]) {
                edges {
                    node {
                        name
                        petKind
                    }
                }
            }
            descSort(sort: [NAME_DESC]) {
                edges {
                    node {
                        name
                    }
                }
            }
            singleColumnSort(sort: NAME_DESC) {
                edges {
                    node {
                        name
                    }
                }
            }
            noDefaultSort(sort: NAME_ASC) {
                edges {
                    node {
                        name
                    }
                }
            }
        }
    """

    def makeNodes(nodeList):
        nodes = [{"node": item} for item in nodeList]
        return {"edges": nodes}

    expected = {
        "defaultSort": makeNodes(
            [
                {"name": "Lassie"},
                {"name": "Barf"},
                {"name": "Alf"},
            ]
        ),
        "nameSort": makeNodes(
            [
                {"name": "Alf"},
                {"name": "Barf"},
                {"name": "Lassie"},
            ]
        ),
        "noDefaultSort": makeNodes(
            [
                {"name": "Alf"},
                {"name": "Barf"},
                {"name": "Lassie"},
            ]
        ),
        "multipleSort": makeNodes(
            [
                {"name": "Alf", "petKind": "CAT"},
                {"name": "Lassie", "petKind": "DOG"},
                {"name": "Barf", "petKind": "DOG"},
            ]
        ),
        "descSort": makeNodes(
            [
                {"name": "Lassie"},
                {"name": "Barf"},
                {"name": "Alf"},
            ]
        ),
        "singleColumnSort": makeNodes(
            [
                {"name": "Lassie"},
                {"name": "Barf"},
                {"name": "Alf"},
            ]
        ),
    }

    schema = Schema(query=Query)
    result = await schema.execute_async(
        query,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Pet]),
        ],
    )
    assert not result.errors
    result = to_std_dicts(result.data)
    assert result == expected


@pytest.mark.asyncio
async def test_sort_query_nulls(session):
    q = sa.insert(Pet).values(
        [
            {"id": 1, "name": "Lassie", "pet_kind": "dog", "hair_kind": HairKind.LONG},
            {"id": 2, "name": "Barf", "pet_kind": "dog", "hair_kind": HairKind.LONG},
            {"id": 3, "name": "Alf", "pet_kind": "cat", "hair_kind": HairKind.LONG},
            {"id": 4, "name": None, "pet_kind": "cat", "hair_kind": HairKind.LONG},
        ]
    )
    await session.execute(q)

    class PetNode(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            interfaces = (Node,)

    class Query(ObjectType):
        defaultSort = SQLAlchemyConnectionField(PetNode.connection)
        nameSort = SQLAlchemyConnectionField(PetNode.connection)
        multipleSort = SQLAlchemyConnectionField(PetNode.connection)
        descSort = SQLAlchemyConnectionField(PetNode.connection)
        singleColumnSort = SQLAlchemyConnectionField(
            PetNode.connection, sort=Argument(PetNode.sort_enum())
        )
        noDefaultSort = SQLAlchemyConnectionField(
            PetNode.connection, sort=PetNode.sort_argument(has_default=False)
        )
        noSort = SQLAlchemyConnectionField(PetNode.connection, sort=None)

    query = """
        query sortTest {
            defaultSort {
                edges {
                    node {
                        name
                    }
                }
            }
            nameSort(sort: NAME_ASC) {
                edges {
                    node {
                        name
                    }
                }
            }
            multipleSort(sort: [PET_KIND_ASC, NAME_DESC]) {
                edges {
                    node {
                        name
                        petKind
                    }
                }
            }
            descSort(sort: [NAME_DESC]) {
                edges {
                    node {
                        name
                    }
                }
            }
            singleColumnSort(sort: NAME_DESC) {
                edges {
                    node {
                        name
                    }
                }
            }
            noDefaultSort(sort: NAME_ASC) {
                edges {
                    node {
                        name
                    }
                }
            }
        }
    """

    def makeNodes(nodeList):
        nodes = [{"node": item} for item in nodeList]
        return {"edges": nodes}

    expected = {
        "defaultSort": makeNodes(
            [
                {"name": "Lassie"},
                {"name": "Barf"},
                {"name": "Alf"},
                {"name": None},
            ]
        ),
        "nameSort": makeNodes(
            [
                {"name": "Alf"},
                {"name": "Barf"},
                {"name": "Lassie"},
                {"name": None},
            ]
        ),
        "noDefaultSort": makeNodes(
            [
                {"name": "Alf"},
                {"name": "Barf"},
                {"name": "Lassie"},
                {"name": None},
            ]
        ),
        "multipleSort": makeNodes(
            [
                {"name": "Alf", "petKind": "CAT"},
                {"name": None, "petKind": "CAT"},
                {"name": "Lassie", "petKind": "DOG"},
                {"name": "Barf", "petKind": "DOG"},
            ]
        ),
        "descSort": makeNodes(
            [
                {"name": "Lassie"},
                {"name": "Barf"},
                {"name": "Alf"},
                {"name": None},
            ]
        ),
        "singleColumnSort": makeNodes(
            [
                {"name": "Lassie"},
                {"name": "Barf"},
                {"name": "Alf"},
                {"name": None},
            ]
        ),
    }

    schema = Schema(query=Query)
    result = await schema.execute_async(
        query,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Pet]),
        ],
    )
    assert not result.errors
    result = to_std_dicts(result.data)
    assert result == expected



import json
import logging
from inspect import isawaitable
from typing import Callable

from graphene import ResolveInfo


class BaseDebugMiddleware:
    def __init__(self, logger, level):
        self.logger = logger
        self.level = level

    async def resolve(self, next_, root, info: ResolveInfo, **args):
        if root is None:
            try:
                self.log(info)
            except Exception as e:
                ...

        result = next_(root, info, **args)
        if isawaitable(result):
            return await result

        return result

    def log(self, info: ResolveInfo):
        raise NotImplementedError()


class LogMiddleware(BaseDebugMiddleware):
    def __init__(
        self,
        logger: logging.Logger = logging.getLogger("gsc"),
        level: int = logging.INFO,
    ):
        super().__init__(logger, level)

    def log(self, info):
        full_query = json.loads(info.context.request._body)
        if full_query.get("operationName") != "IntrospectionQuery":
            text = json.dumps(full_query, ensure_ascii=False, sort_keys=True)
            self.logger.log(self.level, text)


class BreadcrumbMiddleware(BaseDebugMiddleware):
    logger: Callable

    def __init__(self, level: str = "info"):
        from sentry_sdk import add_breadcrumb

        super().__init__(add_breadcrumb, level)

    def log(self, info):
        full_query = json.loads(info.context.request._body)
        if full_query.get("operationName") != "IntrospectionQuery":
            if full_query.get("query"):
                self.logger(
                    category="graphql", message=full_query["query"], level="info"
                )
            if full_query.get("variables"):
                self.logger(
                    category="graphql",
                    message=json.dumps(full_query["variables"]),
                    level="info",
                )



from collections import defaultdict
from typing import TYPE_CHECKING, Type, Union

import sqlalchemy as sa
from graphene import Enum
from sqlalchemy import Column, Table
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.types import Enum as SQLAlchemyEnumType

if TYPE_CHECKING:
    from .types import SQLAlchemyObjectType


class Registry:
    def __init__(self):
        self._registry = defaultdict(list)
        self._registry_models = {}
        self._registry_orm_fields = defaultdict(dict)
        self._registry_composites = {}
        self._registry_enums = {}
        self._registry_sort_enums = {}

    def register(self, obj_type: Type["SQLAlchemyObjectType"]):
        from .types import SQLAlchemyObjectType

        if not isinstance(obj_type, type) or not issubclass(
            obj_type, SQLAlchemyObjectType
        ):
            raise TypeError(f"Expected SQLAlchemyObjectType, but got: {obj_type!r}")
        assert obj_type._meta.registry == self, "Registry for a Model have to match."
        table = sa.inspect(obj_type._meta.model).local_table
        self._registry[table].append(obj_type)

    def get_type_for_model(self, model: Union[DeclarativeMeta, Table], cls_name=None):
        if isinstance(model, DeclarativeMeta):
            model = sa.inspect(model).local_table

        for i in self._registry[model]:
            if cls_name is None or cls_name == i.__name__:
                return i

    def register_orm_field(
        self,
        obj_type: Type["SQLAlchemyObjectType"],
        field_name: str,
        orm_field: Column,
    ):
        from .types import SQLAlchemyObjectType

        if not isinstance(obj_type, type) or not issubclass(
            obj_type, SQLAlchemyObjectType
        ):
            raise TypeError(f"Expected SQLAlchemyObjectType, but got: {obj_type!r}")
        if not field_name or not isinstance(field_name, str):
            raise TypeError(f"Expected a field name, but got: {field_name!r}")
        self._registry_orm_fields[obj_type][field_name] = orm_field

    def get_orm_field_for_graphene_field(
        self, obj_type: Type["SQLAlchemyObjectType"], field_name: str
    ):
        return self._registry_orm_fields.get(obj_type, {}).get(field_name)

    def register_composite_converter(self, composite, converter):
        self._registry_composites[composite] = converter

    def get_converter_for_composite(self, composite):
        return self._registry_composites.get(composite)

    def register_enum(self, sa_enum: SQLAlchemyEnumType, graphene_enum: Enum):
        if not isinstance(sa_enum, SQLAlchemyEnumType):
            raise TypeError(f"Expected SQLAlchemyEnumType, but got: {sa_enum!r}")
        if not isinstance(graphene_enum, type(Enum)):
            raise TypeError(f"Expected Graphene Enum, but got: {graphene_enum!r}")

        self._registry_enums[sa_enum] = graphene_enum

    def get_graphene_enum_for_sa_enum(self, sa_enum: SQLAlchemyEnumType):
        return self._registry_enums.get(sa_enum)

    def register_sort_enum(
        self, obj_type: Type["SQLAlchemyObjectType"], sort_enum: Enum
    ):
        from .types import SQLAlchemyObjectType

        if not isinstance(obj_type, type) or not issubclass(
            obj_type, SQLAlchemyObjectType
        ):
            raise TypeError(f"Expected SQLAlchemyObjectType, but got: {obj_type!r}")
        if not isinstance(sort_enum, type(Enum)):
            raise TypeError(f"Expected Graphene Enum, but got: {sort_enum!r}")
        self._registry_sort_enums[obj_type] = sort_enum

    def get_sort_enum_for_object_type(self, obj_type: Type["SQLAlchemyObjectType"]):
        return self._registry_sort_enums.get(obj_type)


registry = None


def get_global_registry():
    global registry
    if not registry:
        registry = Registry()
    return registry


def reset_global_registry():
    global registry
    registry = None



import pytest
from graphene import Context, ObjectType, Schema, String

from alchql.gql_fields import camel_to_snake
from alchql.utils import (
    to_enum_value_name,
    to_type_name,
)


@pytest.mark.asyncio
async def test_get_session():
    session = "My SQLAlchemy session"

    class Query(ObjectType):
        x = String()

        def resolve_x(self, info):
            return info.context.session

    query = """
        query ReporterQuery {
            x
        }
    """

    schema = Schema(query=Query)
    result = await schema.execute_async(
        query,
        context_value=Context(session=session),
    )
    assert not result.errors
    assert result.data["x"] == session


def test_to_type_name():
    assert to_type_name("make_camel_case") == "MakeCamelCase"
    assert to_type_name("AlreadyCamelCase") == "AlreadyCamelCase"
    assert to_type_name("A_Snake_and_a_Camel") == "ASnakeAndACamel"


def test_camel_to_snake():
    assert camel_to_snake("howYouLikeThat_Eq") == "how_you_like_that__eq"
    assert camel_to_snake("howyoulikethat") == "howyoulikethat"
    assert camel_to_snake("howYouLike000that") == "how_you_like_000that"
    assert camel_to_snake("h000YouLikeThat") == "h000_you_like_that"
    assert camel_to_snake("artistId_Eq") == "artist_id__eq"
    assert camel_to_snake("sort") == "sort"
    assert camel_to_snake("priceMomentum12mo") == "price_momentum_12mo"
    assert camel_to_snake("s3PresignedUrl") == "s3_presigned_url"
    assert camel_to_snake("thumbnailS3PresignedUrl") == "thumbnail_s3_presigned_url"


def test_to_enum_value_name():
    assert to_enum_value_name("make_enum_value_name") == "MAKE_ENUM_VALUE_NAME"
    assert to_enum_value_name("makeEnumValueName") == "MAKE_ENUM_VALUE_NAME"
    assert to_enum_value_name("HTTPStatus400Message") == "HTTP_STATUS400_MESSAGE"
    assert to_enum_value_name("ALREADY_ENUM_VALUE_NAME") == "ALREADY_ENUM_VALUE_NAME"



from unittest.mock import patch

import graphene
import pytest
import sqlalchemy as sa
from graphene import Context

from alchql.connection import from_query
from alchql.connection.utils import offset_to_cursor
from alchql.consts import OP_ILIKE
from alchql.fields import FilterConnectionField
from alchql.middlewares import LoaderMiddleware
from alchql.node import AsyncNode
from alchql.types import SQLAlchemyObjectType
from .models import Editor


async def add_test_data(session):
    await session.execute(
        sa.insert(Editor).values(
            [
                {
                    Editor.name: f"Editor#{num}",
                }
                for num in range(100)
            ]
        )
    )


async def get_query():
    class EditorType(SQLAlchemyObjectType):
        class Meta:
            model = Editor
            interfaces = (AsyncNode,)
            filter_fields = {
                Editor.name: [OP_ILIKE],
            }

    class Query(graphene.ObjectType):
        node = graphene.relay.Node.Field()
        editors = FilterConnectionField(EditorType, sort=EditorType.sort_argument())

    return Query


@pytest.mark.asyncio
async def test_query_forward(session, raise_graphql):
    await add_test_data(session)
    cursor = offset_to_cursor(3)

    query = """
    query {
      editors (first: 3, after:"%s") {
        edges {
          node {
            name
          }
        }
        pageInfo {
          startCursor
          endCursor
          hasPreviousPage
          hasNextPage
        }
      }
    }
    """ % (
        cursor
    )

    schema = graphene.Schema(query=await get_query())
    with patch.object(
        from_query, "get_count_query", side_effect=from_query.get_count_query
    ) as f:
        result = await schema.execute_async(
            query,
            context_value=Context(session=session),
            middleware=[
                LoaderMiddleware([Editor]),
            ],
        )
        f.assert_not_called()

    assert not result.errors
    assert result.data == {
        "editors": {
            "edges": [
                {"node": {"name": "Editor#4"}},
                {"node": {"name": "Editor#5"}},
                {"node": {"name": "Editor#6"}},
            ],
            "pageInfo": {
                "endCursor": "YXJyYXljb25uZWN0aW9uOjY=",
                "hasNextPage": True,
                "hasPreviousPage": True,
                "startCursor": "YXJyYXljb25uZWN0aW9uOjQ=",
            },
        }
    }


@pytest.mark.asyncio
async def test_query_backward(session, raise_graphql):
    await add_test_data(session)
    cursor = offset_to_cursor(3)

    query = """
    query {
      editors (last: 3, before:"%s") {
        edges {
          node {
            name
          }
        }
        pageInfo {
          startCursor
          endCursor
          hasPreviousPage
          hasNextPage
        }
      }
    }
    """ % (
        cursor
    )

    schema = graphene.Schema(query=await get_query())
    with patch.object(
        from_query, "get_count_query", side_effect=from_query.get_count_query
    ) as f:
        result = await schema.execute_async(
            query,
            context_value=Context(session=session),
            middleware=[
                LoaderMiddleware([Editor]),
            ],
        )
        f.assert_not_called()

    assert not result.errors
    assert result.data == {
        "editors": {
            "edges": [
                {"node": {"name": "Editor#0"}},
                {"node": {"name": "Editor#1"}},
                {"node": {"name": "Editor#2"}},
            ],
            "pageInfo": {
                "endCursor": "YXJyYXljb25uZWN0aW9uOjI=",
                "hasNextPage": True,
                "hasPreviousPage": False,
                "startCursor": "YXJyYXljb25uZWN0aW9uOjA=",
            },
        }
    }


@pytest.mark.asyncio
async def test_query_slice(session, raise_graphql):
    await add_test_data(session)
    after = offset_to_cursor(3)
    before = offset_to_cursor(7)

    query = """
    query {
      editors (after: "%s", before:"%s") {
        edges {
          node {
            name
          }
        }
        pageInfo {
          startCursor
          endCursor
          hasPreviousPage
          hasNextPage
        }
      }
    }
    """ % (
        after,
        before,
    )

    schema = graphene.Schema(query=await get_query())
    with patch.object(
        from_query, "get_count_query", side_effect=from_query.get_count_query
    ) as f:
        result = await schema.execute_async(
            query,
            context_value=Context(session=session),
            middleware=[
                LoaderMiddleware([Editor]),
            ],
        )
        f.assert_not_called()

    assert not result.errors
    assert result.data == {
        "editors": {
            "edges": [
                {"node": {"name": "Editor#4"}},
                {"node": {"name": "Editor#5"}},
                {"node": {"name": "Editor#6"}},
            ],
            "pageInfo": {
                "startCursor": "YXJyYXljb25uZWN0aW9uOjQ=",
                "endCursor": "YXJyYXljb25uZWN0aW9uOjY=",
                "hasPreviousPage": True,
                "hasNextPage": True,
            },
        }
    }



from .fields import SQLAlchemyConnectionField
from .sql_mutation import (
    SQLAlchemyUpdateMutation,
    SQLAlchemyCreateMutation,
    SQLAlchemyDeleteMutation,
)
from .types import SQLAlchemyObjectType
from .utils import get_query

__version__ = "3.4.8"

__all__ = [
    "__version__",
    "SQLAlchemyObjectType",
    "SQLAlchemyConnectionField",
    "SQLAlchemyUpdateMutation",
    "SQLAlchemyCreateMutation",
    "SQLAlchemyDeleteMutation",
    "get_query",
]



import time

from graphene import ResolveInfo

from .generated.reports_pb2 import Trace
from .utils import (
    error_to_protobuf_error,
    hr_timestamp_to_nanos,
    response_path_as_string,
)


class TraceTreeBuilder:
    def __init__(self):
        self.trace = Trace()
        self.root_node = self.trace.Node()
        self.start_hr_time = None
        self.stopped = False
        root_response_path = response_path_as_string()
        self.nodes = {root_response_path: self.root_node}

    def start_timing(self):
        if self.start_hr_time:
            raise Exception("start_timing called twice!")
        if self.stopped:
            raise Exception("start_timing called after stop_timing!")
        self.trace.start_time.GetCurrentTime()
        self.start_hr_time = self.trace.start_time

    def stop_timing(self):
        if not self.start_hr_time:
            raise Exception("stop_timing called before start_timing!")
        if self.stopped:
            raise Exception("stop_timing called twice!")
        self.trace.end_time.GetCurrentTime()
        self.trace.duration_ns = hr_timestamp_to_nanos(
            self.trace.end_time
        ) - hr_timestamp_to_nanos(self.trace.start_time)
        self.stopped = True

    def will_resolve_field(self, info: ResolveInfo):
        if not self.start_hr_time:
            raise Exception("will_resolve_field called before start_timing!")
        if self.stopped:
            raise Exception("will_resolve_field called after stop_timing!")
        path = info.path
        node = self.new_node(path)
        node.type = str(info.return_type)
        node.parent_type = str(info.parent_type)
        trace_start_time = hr_timestamp_to_nanos(self.trace.start_time)
        node.start_time = time.time_ns() - trace_start_time
        if type(path.key) == str and path.key != info.field_name:
            # This field was aliased; send the original field name too (for FieldStats).
            node.original_field_name = info.field_name

        def end_node_trace():
            node.end_time = time.time_ns() - trace_start_time

        return end_node_trace

    def did_encounter_errors(self, errors, context):
        for error in errors:
            # @TODO Add more handling here like Apollo's `TraceTreeBuilder`?
            # However, the service name isn't available in the error extensions
            # See: https://github.com/apollographql/apollo-server/blob/b7a91df76acef748488eedcfe998917173cff142/packages/apollo-server-core/src/plugin/traceTreeBuilder.ts#L79
            # @TODO Add support for rewriting errors
            # See: https://github.com/apollographql/apollo-server/blob/b7a91df76acef748488eedcfe998917173cff142/packages/apollo-server-core/src/plugin/traceTreeBuilder.ts#L95
            self.add_protobuf_error(error.path, error_to_protobuf_error(error))

    def new_node(self, path):
        parent_node = self.ensure_parent_node(path)
        node = parent_node.child.add()
        id_ = path.key
        if isinstance(id_, int):
            node.index = id_
        else:
            node.response_name = id_
        response_path = response_path_as_string(path)
        self.nodes[response_path] = node
        return node

    def ensure_parent_node(self, path):
        parent_path = response_path_as_string(path.prev)
        parent_node = self.nodes.get(parent_path, None)
        if parent_node:
            return parent_node

        # Because we set up the root path when creating self.nodes, we now know
        # that path.prev isn't undefined.
        return self.new_node(path.prev)

    def add_protobuf_error(self, path, error):
        if not self.start_hr_time:
            raise Exception("add_protobuf_error called before start_timing!")
        if self.stopped:
            raise Exception("add_protobuf_error called after stop_timing!")
        # By default, put errors on the root node.
        node = self.root_node
        # If a non-GraphQLError Error sneaks in here somehow with a non-array
        # path, don't crash.
        if isinstance(path, list):
            node_key = ".".join(path)
            specified_node = self.nodes.get(node_key, None)
            if specified_node:
                node = specified_node
            else:
                print(
                    f"Could not find node with path {node_key}; "
                    f"defaulting to put errors on root node."
                )
        node.error.append(error)

    def add_nodes_to_trace(self):
        self.trace.root.CopyFrom(self.root_node)



from enum import Enum as PyEnum

import pytest
from graphene import Enum
from sqlalchemy.types import Enum as SQLAlchemyEnumType

from .models import HairKind, Pet
from alchql.enums import _convert_sa_to_graphene_enum, enum_for_field
from alchql.types import SQLAlchemyObjectType


def test_convert_sa_to_graphene_enum_bad_type():
    re_err = "Expected sqlalchemy.types.Enum, but got: 'foo'"
    with pytest.raises(TypeError, match=re_err):
        _convert_sa_to_graphene_enum("foo")


def test_convert_sa_to_graphene_enum_based_on_py_enum():
    class Color(PyEnum):
        RED = 1
        GREEN = 2
        BLUE = 3

    sa_enum = SQLAlchemyEnumType(Color)
    graphene_enum = _convert_sa_to_graphene_enum(sa_enum, "FallbackName")
    assert isinstance(graphene_enum, type(Enum))
    assert graphene_enum._meta.name == "Color"
    assert graphene_enum._meta.enum is Color


def test_convert_sa_to_graphene_enum_based_on_py_enum_with_bad_names():
    class Color(PyEnum):
        red = 1
        green = 2
        blue = 3

    sa_enum = SQLAlchemyEnumType(Color)
    graphene_enum = _convert_sa_to_graphene_enum(sa_enum, "FallbackName")
    assert isinstance(graphene_enum, type(Enum))
    assert graphene_enum._meta.name == "Color"
    assert graphene_enum._meta.enum is not Color
    assert [
        (key, value.value)
        for key, value in graphene_enum._meta.enum.__members__.items()
    ] == [("RED", 1), ("GREEN", 2), ("BLUE", 3)]


def test_convert_sa_enum_to_graphene_enum_based_on_list_named():
    sa_enum = SQLAlchemyEnumType("red", "green", "blue", name="color_values")
    graphene_enum = _convert_sa_to_graphene_enum(sa_enum, "FallbackName")
    assert isinstance(graphene_enum, type(Enum))
    assert graphene_enum._meta.name == "ColorValues"
    assert [
        (key, value.value)
        for key, value in graphene_enum._meta.enum.__members__.items()
    ] == [("RED", "red"), ("GREEN", "green"), ("BLUE", "blue")]


def test_convert_sa_enum_to_graphene_enum_based_on_list_unnamed():
    sa_enum = SQLAlchemyEnumType("red", "green", "blue")
    graphene_enum = _convert_sa_to_graphene_enum(sa_enum, "FallbackName")
    assert isinstance(graphene_enum, type(Enum))
    assert graphene_enum._meta.name == "FallbackName"
    assert [
        (key, value.value)
        for key, value in graphene_enum._meta.enum.__members__.items()
    ] == [("RED", "red"), ("GREEN", "green"), ("BLUE", "blue")]


def test_convert_sa_enum_to_graphene_enum_based_on_list_without_name():
    sa_enum = SQLAlchemyEnumType("red", "green", "blue")
    re_err = r"No type name specified for Enum\('red', 'green', 'blue'\)"
    with pytest.raises(TypeError, match=re_err):
        _convert_sa_to_graphene_enum(sa_enum)


def test_enum_for_field():
    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet

    enum = enum_for_field(PetType, "pet_kind")
    assert isinstance(enum, type(Enum))
    assert enum._meta.name == "PetKind"
    assert [
        (key, value.value) for key, value in enum._meta.enum.__members__.items()
    ] == [("CAT", "cat"), ("DOG", "dog")]
    enum2 = enum_for_field(PetType, "pet_kind")
    assert enum2 is enum
    enum2 = PetType.enum_for_field("pet_kind")
    assert enum2 is enum

    enum = enum_for_field(PetType, "hair_kind")
    assert isinstance(enum, type(Enum))
    assert enum._meta.name == "HairKind"
    assert enum._meta.enum is HairKind
    enum2 = PetType.enum_for_field("hair_kind")
    assert enum2 is enum

    re_err = r"Cannot get PetType\.other_kind"
    with pytest.raises(TypeError, match=re_err):
        enum_for_field(PetType, "other_kind")
    with pytest.raises(TypeError, match=re_err):
        PetType.enum_for_field("other_kind")

    re_err = r"PetType\.name does not map to enum column"
    with pytest.raises(TypeError, match=re_err):
        enum_for_field(PetType, "name")
    with pytest.raises(TypeError, match=re_err):
        PetType.enum_for_field("name")

    re_err = r"Expected a field name, but got: None"
    with pytest.raises(TypeError, match=re_err):
        enum_for_field(PetType, None)
    with pytest.raises(TypeError, match=re_err):
        PetType.enum_for_field(None)

    re_err = "Expected SQLAlchemyObjectType, but got: None"
    with pytest.raises(TypeError, match=re_err):
        enum_for_field(None, "other_kind")



from unittest import mock

import pytest
import sqlalchemy as sa
from graphene import (
    Context,
    Dynamic,
    Field,
    GlobalID,
    Int,
    List,
    Node,
    NonNull,
    ObjectType,
    Schema,
    String,
)
from graphene.relay import Connection
from sqlalchemy import Column, ForeignKey, Integer
from sqlalchemy.orm import declarative_base, relationship

from alchql import gql_types
from alchql.converter import convert_sqlalchemy_composite
from alchql.fields import (
    BatchSQLAlchemyConnectionField,
    SQLAlchemyConnectionField,
    UnsortedSQLAlchemyConnectionField,
)
from alchql.middlewares import LoaderMiddleware
from alchql.node import AsyncNode
from alchql.types import (
    ORMField,
    SQLAlchemyObjectType,
    SQLAlchemyObjectTypeOptions,
)
from .models import Article, CompositeFullName, Pet, Reporter


def test_should_raise_if_no_model():
    re_err = r"valid SQLAlchemy Model"
    with pytest.raises(Exception, match=re_err):

        class Character1(SQLAlchemyObjectType):
            pass


def test_should_raise_if_model_is_invalid():
    re_err = r"valid SQLAlchemy Model"
    with pytest.raises(Exception, match=re_err):

        class Character(SQLAlchemyObjectType):
            class Meta:
                model = 1


@pytest.mark.asyncio
async def test_sqlalchemy_node(session):
    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter
            interfaces = (Node,)

    reporter_id_field = ReporterType._meta.fields["id"]
    assert isinstance(reporter_id_field, GlobalID)

    reporter_id = (await session.execute(sa.insert(Reporter))).lastrowid
    reporter: Reporter = (
        await session.execute(
            sa.select(Reporter.__table__.c).where(Reporter.id == reporter_id)
        )
    ).first()

    info = mock.Mock(context=type("Context", (), {"session": session}))
    reporter_node = await ReporterType.get_node(info, reporter.id)
    assert reporter.id == reporter_node.id
    assert reporter.first_name == reporter_node.first_name
    assert reporter.last_name == reporter_node.last_name
    assert reporter.email == reporter_node.email
    assert reporter.favorite_pet_kind == reporter_node.favorite_pet_kind


@pytest.mark.asyncio
async def test_sqlalchemy_node_with_exclude(session):
    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter
            interfaces = (Node,)
            exclude_fields = ["first_name", "last_name"]

        other = gql_types.String()

    reporter_id_field = ReporterType._meta.fields["id"]
    assert isinstance(reporter_id_field, GlobalID)

    reporter_id = (await session.execute(sa.insert(Reporter))).lastrowid
    reporter: Reporter = (
        await session.execute(
            sa.select(Reporter.__table__.c).where(Reporter.id == reporter_id)
        )
    ).first()

    info = mock.Mock(context=type("Context", (), {"session": session}))
    reporter_node = await ReporterType.get_node(info, reporter.id)
    assert reporter.id == reporter_node.id
    assert reporter.email == reporter_node.email
    assert reporter.favorite_pet_kind == reporter_node.favorite_pet_kind
    assert reporter_node.other is None


def test_connection():
    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter
            interfaces = (Node,)

    assert issubclass(ReporterType.connection, Connection)


def test_sqlalchemy_default_fields():
    @convert_sqlalchemy_composite.register(CompositeFullName)
    def convert_composite_class(composite, registry):
        return String()

    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter
            interfaces = (Node,)

    class ArticleType(SQLAlchemyObjectType):
        class Meta:
            model = Article
            interfaces = (Node,)

    assert set(ReporterType._meta.fields.keys()) == {
        # Columns
        "column_prop",  # SQLAlchemy retuns column properties first
        "id",
        "first_name",
        "last_name",
        "email",
        "favorite_pet_kind",
        # Composite
        "composite_prop",
        # Hybrid
        "hybrid_prop",
        # Relationship
        "pets",
        "articles",
        "favorite_article",
    }

    # column
    first_name_field = ReporterType._meta.fields["first_name"]
    assert first_name_field.type == String
    assert first_name_field.description == "First name"

    # column_property
    column_prop_field = ReporterType._meta.fields["column_prop"]
    assert column_prop_field.type == Int
    # "doc" is ignored by column_property
    assert column_prop_field.description is None

    # composite
    full_name_field = ReporterType._meta.fields["composite_prop"]
    assert full_name_field.type == String
    # "doc" is ignored by composite
    assert full_name_field.description is None

    # hybrid_property
    hybrid_prop = ReporterType._meta.fields["hybrid_prop"]
    assert hybrid_prop.type == String
    # "doc" is ignored by hybrid_property
    assert hybrid_prop.description is None

    # relationship
    favorite_article_field = ReporterType._meta.fields["favorite_article"]
    assert isinstance(favorite_article_field, Dynamic)
    assert favorite_article_field.type().type == ArticleType
    assert favorite_article_field.type().description is None


def test_sqlalchemy_override_fields():
    @convert_sqlalchemy_composite.register(CompositeFullName)
    def convert_composite_class(composite, registry):
        return String()

    class ReporterMixin(object):
        # columns
        first_name = ORMField(required=True)
        last_name = ORMField(description="Overridden")

    class ReporterType(SQLAlchemyObjectType, ReporterMixin):
        class Meta:
            model = Reporter
            interfaces = (Node,)

        # columns
        email = ORMField(deprecation_reason="Overridden")
        email_v2 = ORMField(model_attr="email", type_=Int)

        # column_property
        column_prop = ORMField(type_=String)

        # composite
        composite_prop = ORMField()

        # hybrid_property
        hybrid_prop = ORMField(description="Overridden")

        # relationships
        favorite_article = ORMField(description="Overridden")
        pets = ORMField(description="Overridden")

    class ArticleType(SQLAlchemyObjectType):
        class Meta:
            model = Article
            interfaces = (Node,)

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            interfaces = (Node,)
            use_connection = False

    assert set(ReporterType._meta.fields.keys()) == {
        # Fields from ReporterMixin
        "first_name",
        "last_name",
        # Fields from ReporterType
        "email",
        "email_v2",
        "column_prop",
        "composite_prop",
        "hybrid_prop",
        "favorite_article",
        "pets",
        "articles",
        # Then the automatic SQLAlchemy fields
        "id",
        "favorite_pet_kind",
    }

    first_name_field = ReporterType._meta.fields["first_name"]
    assert isinstance(first_name_field.type, NonNull)
    assert first_name_field.type.of_type == String
    assert first_name_field.description == "First name"
    assert first_name_field.deprecation_reason is None

    last_name_field = ReporterType._meta.fields["last_name"]
    assert last_name_field.type == String
    assert last_name_field.description == "Overridden"
    assert last_name_field.deprecation_reason is None

    email_field = ReporterType._meta.fields["email"]
    assert email_field.type == String
    assert email_field.description == "Email"
    assert email_field.deprecation_reason == "Overridden"

    email_field_v2 = ReporterType._meta.fields["email_v2"]
    assert email_field_v2.type == Int
    assert email_field_v2.description == "Email"
    assert email_field_v2.deprecation_reason is None

    hybrid_prop_field = ReporterType._meta.fields["hybrid_prop"]
    assert hybrid_prop_field.type == String
    assert hybrid_prop_field.description == "Overridden"
    assert hybrid_prop_field.deprecation_reason is None

    column_prop_field_v2 = ReporterType._meta.fields["column_prop"]
    assert column_prop_field_v2.type == String
    assert column_prop_field_v2.description is None
    assert column_prop_field_v2.deprecation_reason is None

    composite_prop_field = ReporterType._meta.fields["composite_prop"]
    assert composite_prop_field.type == String
    assert composite_prop_field.description is None
    assert composite_prop_field.deprecation_reason is None

    favorite_article_field = ReporterType._meta.fields["favorite_article"]
    assert isinstance(favorite_article_field, Dynamic)
    assert favorite_article_field.type().type == ArticleType
    assert favorite_article_field.type().description == "Overridden"

    articles_field = ReporterType._meta.fields["articles"]
    assert isinstance(articles_field, Dynamic)
    assert isinstance(articles_field.type(), UnsortedSQLAlchemyConnectionField)
    # assert articles_field.type().deprecation_reason == "Overridden"

    pets_field = ReporterType._meta.fields["pets"]
    assert isinstance(pets_field, Dynamic)
    assert isinstance(pets_field.type().type, List)
    assert pets_field.type().type.of_type == PetType
    assert pets_field.type().description == "Overridden"


def test_invalid_model_attr():
    err_msg = (
        "Cannot map ORMField to a model attribute.\n" "Field: 'ReporterType.first_name'"
    )
    with pytest.raises(ValueError, match=err_msg):

        class ReporterType(SQLAlchemyObjectType):
            class Meta:
                model = Reporter

            first_name = ORMField(model_attr="does_not_exist")


def test_only_fields():
    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter
            only_fields = ("id", "last_name")

        first_name = ORMField()  # Takes precedence
        last_name = ORMField()  # Noop

    assert list(ReporterType._meta.fields.keys()) == ["first_name", "last_name", "id"]


def test_exclude_fields():
    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter
            exclude_fields = ("id", "first_name")

        first_name = ORMField()  # Takes precedence
        last_name = ORMField()  # Noop

    assert set(ReporterType._meta.fields.keys()) == {
        "first_name",
        "last_name",
        "column_prop",
        "email",
        "favorite_pet_kind",
        "composite_prop",
        "hybrid_prop",
        "pets",
        "articles",
        "favorite_article",
    }


def test_only_and_exclude_fields():
    re_err = r"'only_fields' and 'exclude_fields' cannot be both set"
    with pytest.raises(Exception, match=re_err):

        class ReporterType(SQLAlchemyObjectType):
            class Meta:
                model = Reporter
                only_fields = ("id", "last_name")
                exclude_fields = ("id", "last_name")


def test_sqlalchemy_redefine_field():
    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter

        first_name = Int()

    first_name_field = ReporterType._meta.fields["first_name"]
    assert isinstance(first_name_field, Field)
    assert first_name_field.type == Int


def test_sqlalchemy_redefine_field2():
    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter

        first_name = gql_types.Int(model_field=Reporter.first_name)

    first_name_field = ReporterType._meta.fields["first_name"]
    assert isinstance(first_name_field, Field)
    assert first_name_field.type == gql_types.Int
    assert first_name_field.type != Int


@pytest.mark.asyncio
async def test_sqlalchemy_redefine_field3(session):
    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter
            connection_field_factory = BatchSQLAlchemyConnectionField.from_relationship
            interfaces = (AsyncNode,)

        fname = gql_types.String(model_field=Reporter.first_name, name="nameObj")

    class Query(ObjectType):
        reporter = SQLAlchemyConnectionField(ReporterType.connection, sort=None)

    reporter = Reporter(
        first_name="first_name",
        last_name="last_name",
        email="email",
        favorite_pet_kind="cat",
    )
    session.add(reporter)
    await session.commit()

    schema = Schema(query=Query, types=[ReporterType])
    result = await schema.execute_async(
        """
        query {
            reporter {
                edges{
                    node{
                        id
                        nameObj
                    }
                }
            }
        }
        """,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Reporter]),
        ],
    )

    assert not result.errors
    assert result.data["reporter"]["edges"][0]["node"]["nameObj"] == "first_name"


@pytest.mark.asyncio
async def test_sqlalchemy_redefine_field4(session):
    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter
            connection_field_factory = BatchSQLAlchemyConnectionField.from_relationship
            interfaces = (AsyncNode,)

        fname = gql_types.String(model_field=Reporter.first_name)

    class Query(ObjectType):
        reporter = SQLAlchemyConnectionField(ReporterType.connection, sort=None)

    reporter = Reporter(
        first_name="first_name",
        last_name="last_name",
        email="email",
        favorite_pet_kind="cat",
    )
    session.add(reporter)
    await session.commit()

    schema = Schema(query=Query, types=[ReporterType])
    result = await schema.execute_async(
        """
        query {
            reporter {
                edges{
                    node{
                        id
                        fname
                    }
                }
            }
        }
        """,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Reporter]),
        ],
    )

    assert not result.errors
    assert result.data["reporter"]["edges"][0]["node"]["fname"] == "first_name"


@pytest.mark.asyncio
async def test_resolvers(session):
    """Test that the correct resolver functions are called"""

    class ReporterMixin(object):
        def resolve_id(root, _info):
            return "ID"

    class ReporterType(ReporterMixin, SQLAlchemyObjectType):
        class Meta:
            model = Reporter

        email = ORMField()
        email_v2 = ORMField(model_attr="email")
        favorite_pet_kind = Field(String)
        favorite_pet_kind_v2 = Field(String)

        def resolve_last_name(root, _info):
            return root.last_name.upper()

        def resolve_email_v2(root, _info):
            return root.email + "_V2"

        def resolve_favorite_pet_kind_v2(root, _info):
            return str(root.favorite_pet_kind) + "_V2"

    class Query(ObjectType):
        reporter = Field(ReporterType)

        async def resolve_reporter(self, _info):
            return (await session.execute(sa.select(Reporter))).scalars().first()

    reporter = Reporter(
        first_name="first_name",
        last_name="last_name",
        email="email",
        favorite_pet_kind="cat",
    )
    session.add(reporter)
    await session.commit()

    schema = Schema(query=Query)
    result = await schema.execute_async(
        """
        query {
            reporter {
                id
                firstName
                lastName
                email
                emailV2
                favoritePetKind
                favoritePetKindV2
            }
        }
    """
    )

    assert not result.errors
    # Custom resolver on a base class
    assert result.data["reporter"]["id"] == "ID"
    # Default field + default resolver
    assert result.data["reporter"]["firstName"] == "first_name"
    # Default field + custom resolver
    assert result.data["reporter"]["lastName"] == "LAST_NAME"
    # ORMField + default resolver
    assert result.data["reporter"]["email"] == "email"
    # ORMField + custom resolver
    assert result.data["reporter"]["emailV2"] == "email_V2"
    # Field + default resolver
    assert result.data["reporter"]["favoritePetKind"] == "cat"
    # Field + custom resolver
    assert result.data["reporter"]["favoritePetKindV2"] == "cat_V2"


# Test Custom SQLAlchemyObjectType Implementation


def test_custom_objecttype_registered():
    class CustomSQLAlchemyObjectType(SQLAlchemyObjectType):
        class Meta:
            abstract = True

    class CustomReporterType(CustomSQLAlchemyObjectType):
        class Meta:
            model = Reporter

    assert issubclass(CustomReporterType, ObjectType)
    assert CustomReporterType._meta.model == Reporter
    assert len(CustomReporterType._meta.fields) == 11


# Test Custom SQLAlchemyObjectType with Custom Options
def test_objecttype_with_custom_options():
    class CustomOptions(SQLAlchemyObjectTypeOptions):
        custom_option = None

    class SQLAlchemyObjectTypeWithCustomOptions(SQLAlchemyObjectType):
        class Meta:
            abstract = True

        @classmethod
        def __init_subclass_with_meta__(cls, custom_option=None, **options):
            _meta = CustomOptions(cls)
            _meta.custom_option = custom_option
            super().__init_subclass_with_meta__(_meta=_meta, **options)

    class ReporterWithCustomOptions(SQLAlchemyObjectTypeWithCustomOptions):
        class Meta:
            model = Reporter
            custom_option = "custom_option"

    assert issubclass(ReporterWithCustomOptions, ObjectType)
    assert ReporterWithCustomOptions._meta.model == Reporter
    assert ReporterWithCustomOptions._meta.custom_option == "custom_option"


# Tests for connection_field_factory


class _TestSQLAlchemyConnectionField(SQLAlchemyConnectionField):
    pass


def test_default_connection_field_factory():
    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter
            interfaces = (Node,)

    class ArticleType(SQLAlchemyObjectType):
        class Meta:
            model = Article
            interfaces = (Node,)

    assert isinstance(
        ReporterType._meta.fields["articles"].type(), UnsortedSQLAlchemyConnectionField
    )


def test_custom_connection_field_factory():
    _Base = declarative_base()

    def test_connection_field_factory(relationship, registry):
        model = relationship.mapper.entity
        _type = registry.get_type_for_model(model)
        return _TestSQLAlchemyConnectionField(_type._meta.connection)

    class _Article(_Base):
        __tablename__ = "articles"

        id = Column(Integer(), primary_key=True)
        reporter_id = Column(Integer(), ForeignKey("reporters.id"))

    class _Reporter(_Base):
        __tablename__ = "reporters"

        id = Column(Integer(), primary_key=True)
        articles = relationship(_Article, backref="reporter")

    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = _Reporter
            interfaces = (Node,)
            connection_field_factory = test_connection_field_factory

    class ArticleType(SQLAlchemyObjectType):
        class Meta:
            model = _Article
            interfaces = (Node,)

    assert isinstance(
        ReporterType._meta.fields["articles"].type(),
        _TestSQLAlchemyConnectionField,
    )



import graphene
import pytest
import sqlalchemy as sa
from graphene import Context

from .models import Article, Base, CompositeFullName, Editor, HairKind, Pet, Reporter
from .utils import to_std_dicts
from alchql.converter import convert_sqlalchemy_composite
from alchql.fields import SQLAlchemyConnectionField
from alchql.middlewares import LoaderMiddleware
from alchql.node import AsyncNode
from alchql.types import ORMField, SQLAlchemyObjectType


async def add_test_data(session):
    reporter_id = (
        await session.execute(
            sa.insert(Reporter).values(
                {
                    Reporter.first_name: "John",
                    Reporter.last_name: "Doe",
                    Reporter.favorite_pet_kind: "cat",
                }
            )
        )
    ).lastrowid

    await session.execute(
        sa.insert(Pet).values(
            {
                Pet.name: "Garfield",
                Pet.pet_kind: "cat",
                Pet.hair_kind: HairKind.SHORT,
                Pet.reporter_id: reporter_id,
            }
        )
    )

    await session.execute(
        sa.insert(Article).values(
            {
                Article.headline: "Hi!",
                Article.reporter_id: reporter_id,
            }
        )
    )

    reporter_id = (
        await session.execute(
            sa.insert(Reporter).values(
                {
                    Reporter.first_name: "Jane",
                    Reporter.last_name: "Roe",
                    Reporter.favorite_pet_kind: "dog",
                }
            )
        )
    ).lastrowid

    await session.execute(
        sa.insert(Pet).values(
            {
                Pet.name: "Lassie",
                Pet.pet_kind: "dog",
                Pet.hair_kind: HairKind.LONG,
                Pet.reporter_id: reporter_id,
            }
        )
    )

    await session.execute(
        sa.insert(Editor).values(
            {
                Editor.name: "Jack",
            }
        )
    )


@pytest.mark.asyncio
async def test_query_fields(session):
    await add_test_data(session)

    @convert_sqlalchemy_composite.register(CompositeFullName)
    def convert_composite_class(composite, registry):
        return graphene.String()

    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter

    class Query(graphene.ObjectType):
        reporter = graphene.Field(ReporterType)
        reporters = graphene.List(ReporterType)

        async def resolve_reporter(self, _info):
            _result = await session.execute(sa.select(Reporter))
            return _result.scalars().first()

        async def resolve_reporters(self, _info):
            _result = await session.execute(sa.select(Reporter))
            return _result.scalars().all()

    query = """
        query {
          reporter {
            firstName
            columnProp
            hybridProp
            compositeProp
          }
          reporters {
            firstName
          }
        }
    """
    expected = {
        "reporter": {
            "firstName": "John",
            "hybridProp": "John",
            "columnProp": 2,
            "compositeProp": "John Doe",
        },
        "reporters": [
            {"firstName": "John"},
            {"firstName": "Jane"},
        ],
    }
    schema = graphene.Schema(query=Query)
    result = await schema.execute_async(query)
    assert not result.errors
    result = to_std_dicts(result.data)
    assert result == expected


@pytest.mark.asyncio
async def test_query_node(session):
    await add_test_data(session)

    class ReporterNode(SQLAlchemyObjectType):
        class Meta:
            model = Reporter
            interfaces = (AsyncNode,)

        @classmethod
        def get_node(cls, info, id):
            return Reporter(id=2, first_name="Cookie Monster")

    class ArticleNode(SQLAlchemyObjectType):
        class Meta:
            model = Article
            interfaces = (AsyncNode,)

    class Query(graphene.ObjectType):
        node = AsyncNode.Field()
        reporter = graphene.Field(ReporterNode)
        all_articles = SQLAlchemyConnectionField(ArticleNode.connection)

        async def resolve_reporter(self, _info):
            return (await session.execute(sa.select(Reporter))).scalars().first()

    query = """
        query {
          reporter {
            id
            firstName
            articles {
              edges {
                node {
                  headline
                }
              }
            }
          }
          allArticles {
            edges {
              node {
                headline
              }
            }
          }
          myArticle: node(id:"QXJ0aWNsZU5vZGU6MQ==") {
            id
            ... on ReporterNode {
                firstName
            }
            ... on ArticleNode {
                headline
            }
          }
        }
    """
    expected = {
        "reporter": {
            "id": "UmVwb3J0ZXJOb2RlOjE=",
            "firstName": "John",
            "articles": {"edges": [{"node": {"headline": "Hi!"}}]},
        },
        "allArticles": {"edges": [{"node": {"headline": "Hi!"}}]},
        "myArticle": {"id": "QXJ0aWNsZU5vZGU6MQ==", "headline": "Hi!"},
    }
    schema = graphene.Schema(query=Query)
    result = await schema.execute_async(
        query,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Article, Reporter]),
        ],
    )
    assert not result.errors
    result = to_std_dicts(result.data)
    assert result == expected


@pytest.mark.asyncio
async def test_orm_field(session):
    await add_test_data(session)

    @convert_sqlalchemy_composite.register(CompositeFullName)
    def convert_composite_class(composite, registry):
        return graphene.String()

    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter
            interfaces = (AsyncNode,)

        first_name_v2 = ORMField(model_attr="first_name")
        hybrid_prop_v2 = ORMField(model_attr="hybrid_prop")
        column_prop_v2 = ORMField(model_attr="column_prop")
        composite_prop = ORMField()
        favorite_article_v2 = ORMField(model_attr="favorite_article")

    class ArticleType(SQLAlchemyObjectType):
        class Meta:
            model = Article
            interfaces = (AsyncNode,)

    class Query(graphene.ObjectType):
        reporter = graphene.Field(ReporterType)

        async def resolve_reporter(self, _info):
            return (await session.execute(sa.select(Reporter))).scalars().first()

    query = """
        query {
          reporter {
            firstNameV2
            hybridPropV2
            columnPropV2
            compositeProp
            favoriteArticleV2 {
              headline
            }
          }
        }
    """
    expected = {
        "reporter": {
            "firstNameV2": "John",
            "hybridPropV2": "John",
            "columnPropV2": 2,
            "compositeProp": "John Doe",
            "favoriteArticleV2": {"headline": "Hi!"},
        },
    }
    schema = graphene.Schema(query=Query)
    result = await schema.execute_async(
        query,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Article, Reporter]),
        ],
    )
    assert not result.errors
    result = to_std_dicts(result.data)
    assert result == expected


@pytest.mark.asyncio
async def test_custom_identifier(session):
    await add_test_data(session)

    class EditorNode(SQLAlchemyObjectType):
        class Meta:
            model = Editor
            interfaces = (AsyncNode,)

    class Query(graphene.ObjectType):
        node = AsyncNode.Field()
        all_editors = SQLAlchemyConnectionField(EditorNode.connection)

    query = """
        query {
          allEditors {
            edges {
                node {
                    id
                    name
                }
            }
          },
          node(id: "RWRpdG9yTm9kZTox") {
            ...on EditorNode {
              name
            }
          }
        }
    """
    expected = {
        "allEditors": {"edges": [{"node": {"id": "RWRpdG9yTm9kZTox", "name": "Jack"}}]},
        "node": {"name": "Jack"},
    }

    schema = graphene.Schema(query=Query)
    result = await schema.execute_async(
        query,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Article, Reporter]),
        ],
    )
    assert not result.errors
    result = to_std_dicts(result.data)
    assert result == expected


@pytest.mark.asyncio
async def test_mutation(session):
    await add_test_data(session)

    class EditorNode(SQLAlchemyObjectType):
        class Meta:
            model = Editor
            interfaces = (AsyncNode,)

    class ReporterNode(SQLAlchemyObjectType):
        class Meta:
            model = Reporter
            interfaces = (AsyncNode,)

        @classmethod
        async def get_node(cls, id, info):
            return Reporter(id=2, first_name="Cookie Monster")

    class ArticleNode(SQLAlchemyObjectType):
        class Meta:
            model = Article
            interfaces = (AsyncNode,)

    class CreateArticle(graphene.Mutation):
        class Arguments:
            headline = graphene.String()
            reporter_id = graphene.ID()

        ok = graphene.Boolean()
        article = graphene.Field(ArticleNode)

        async def mutate(self, info, headline, reporter_id):
            s = info.context.session
            _result = await s.execute(
                sa.insert(Article).values(
                    dict(headline=headline, reporter_id=reporter_id)
                )
            )
            _result2 = await s.execute(
                sa.select(Article).where(Article.id == _result.lastrowid)
            )
            new_article = _result2.scalars().first()

            return CreateArticle(article=new_article, ok=True)

    class Query(graphene.ObjectType):
        node = AsyncNode.Field()

    class Mutation(graphene.ObjectType):
        create_article = CreateArticle.Field()

    query = """
        mutation {
          createArticle(
            headline: "My Article"
            reporterId: "1"
          ) {
            ok
            article {
                headline
                reporter {
                    id
                    firstName
                }
            }
          }
        }
    """
    expected = {
        "createArticle": {
            "ok": True,
            "article": {
                "headline": "My Article",
                "reporter": {"id": "UmVwb3J0ZXJOb2RlOjE=", "firstName": "John"},
            },
        }
    }

    schema = graphene.Schema(query=Query, mutation=Mutation)
    result = await schema.execute_async(
        query,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware(Base.registry.mappers),
        ],
    )
    assert not result.errors
    result = to_std_dicts(result.data)
    assert result == expected
